#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file    python_isotropic_stiffness_operator_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    31 Dec 2025

@brief   Tests for IsotropicStiffnessOperator (fused elliptic kernel)

Copyright © 2025 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np
import pytest

# Import GPU testing utilities from conftest
from conftest import (
    HAS_CUPY,
    cp,
    create_device,
    get_test_devices,
    skip_if_gpu_unavailable,
)

import muGrid
from muGrid import Timer

# =============================================================================
# Test Configuration
# =============================================================================

GPU_AVAILABLE = muGrid.has_gpu


# =============================================================================
# 2D Operator Tests
# =============================================================================


class TestIsotropicStiffnessOperator2D:
    """Test suite for IsotropicStiffnessOperator2D."""

    def test_construction(self):
        """Test that the operator can be constructed."""
        grid_spacing = [0.1, 0.1]
        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Check that G and V matrices have correct shape
        G = op.G
        V = op.V
        assert G.shape == (8, 8)
        assert V.shape == (8, 8)

    def test_symmetry(self):
        """Test that G and V matrices are symmetric."""
        grid_spacing = [0.1, 0.2]
        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        G = op.G
        V = op.V

        np.testing.assert_allclose(G, G.T, rtol=1e-10)
        np.testing.assert_allclose(V, V.T, rtol=1e-10)

    def test_apply_homogeneous_material(self):
        """Test apply with homogeneous material."""
        nx, ny = 8, 8
        grid_spacing = [1.0 / (nx - 1), 1.0 / (ny - 1)]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Create field collection with ghost cells on both sides
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )

        # Create fields
        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))

        # Create material field collection (same size as node field, with ghosts)
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))

        # Set homogeneous material properties
        lam = 1.0
        mu = 1.0
        lambda_field.p[...] = lam
        mu_field.p[...] = mu

        # Set a simple displacement field (linear in x)
        x = np.arange(nx) * grid_spacing[0]
        y = np.arange(ny) * grid_spacing[1]
        X, Y = np.meshgrid(x, y, indexing="ij")
        displacement.p[0, ...] = X  # u_x = x
        displacement.p[1, ...] = 0  # u_y = 0

        # Fill ghost cells (periodic BC)
        displacement.pg[0, 1 : nx + 1, 1 : ny + 1] = displacement.p[0]
        displacement.pg[1, 1 : nx + 1, 1 : ny + 1] = displacement.p[1]
        displacement.pg[:, 0, :] = displacement.pg[:, nx, :]  # left ghost
        displacement.pg[:, nx + 1, :] = displacement.pg[:, 1, :]  # right ghost
        displacement.pg[:, :, 0] = displacement.pg[:, :, ny]  # bottom ghost
        displacement.pg[:, :, ny + 1] = displacement.pg[:, :, 1]  # top ghost

        # Fill material ghost cells
        lambda_field.pg[..., 1 : nx + 1, 1 : ny + 1] = lambda_field.p
        lambda_field.pg[..., 0, :] = lambda_field.pg[..., nx, :]
        lambda_field.pg[..., nx + 1, :] = lambda_field.pg[..., 1, :]
        lambda_field.pg[..., :, 0] = lambda_field.pg[..., :, ny]
        lambda_field.pg[..., :, ny + 1] = lambda_field.pg[..., :, 1]
        mu_field.pg[..., 1 : nx + 1, 1 : ny + 1] = mu_field.p
        mu_field.pg[..., 0, :] = mu_field.pg[..., nx, :]
        mu_field.pg[..., nx + 1, :] = mu_field.pg[..., 1, :]
        mu_field.pg[..., :, 0] = mu_field.pg[..., :, ny]
        mu_field.pg[..., :, ny + 1] = mu_field.pg[..., :, 1]

        # Apply operator
        op.apply(displacement, lambda_field, mu_field, force)

        # Force should be non-zero only at boundaries for uniform strain
        # (internal forces cancel out for constant strain)
        # Just check that it runs without error and force is finite
        assert np.all(np.isfinite(force.p))

    def test_compare_with_generic_operator(self):
        """Compare fused kernel with generic B^T C B computation."""
        nn = 6  # Number of nodes
        nel = nn  # Number of elements (periodic BC)
        grid_spacing = [0.25, 0.25]

        # Material properties (homogeneous)
        lam = 2.0
        mu = 1.5

        # Create the fused operator
        fused_op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Create the generic gradient operator
        grad_op = muGrid.FEMGradientOperator(2, grid_spacing)

        # Create decomposition with ghosts for periodic BC
        comm = muGrid.Communicator()
        decomposition = muGrid.CartesianDecomposition(
            comm,
            (nn, nn),
            nb_subdivisions=(1, 1),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            nb_sub_pts={"quad": grad_op.nb_quad_pts},
        )

        # Create material field collection with ghost cells for periodic BC
        # (ghost cells are filled via communicate_ghosts once before apply)
        fc_mat = muGrid.GlobalFieldCollection(
            (nel, nel), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = lam
        mu_field.p[...] = mu

        # Fill ghost cells with periodic copies
        # (This is done once if material field doesn't change during simulation)
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(mu_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        # Create displacement field
        displacement = decomposition.real_field("displacement", (2,))
        np.random.seed(42)
        displacement.p[...] = np.random.rand(*displacement.p.shape)
        decomposition.communicate_ghosts(displacement)

        # Force fields
        force_fused = decomposition.real_field("force_fused", (2,))
        force_generic = decomposition.real_field("force_generic", (2,))

        # Apply fused operator
        fused_op.apply(displacement, lambda_field, mu_field, force_fused)

        # Apply generic operator (gradient -> stress -> divergence)
        gradient = decomposition.real_field("gradient", (2, 2), "quad")
        grad_op.apply(displacement, gradient)

        # Compute stress using isotropic constitutive law
        stress = decomposition.real_field("stress", (2, 2), "quad")
        grad = gradient.s

        # Symmetric strain
        eps_xx = grad[0, 0, ...]
        eps_yy = grad[1, 1, ...]
        eps_xy = 0.5 * (grad[0, 1, ...] + grad[1, 0, ...])
        trace = eps_xx + eps_yy

        # Stress
        stress.s[0, 0, ...] = lam * trace + 2 * mu * eps_xx
        stress.s[1, 1, ...] = lam * trace + 2 * mu * eps_yy
        stress.s[0, 1, ...] = 2 * mu * eps_xy
        stress.s[1, 0, ...] = 2 * mu * eps_xy

        # Fill ghost cells for periodic contribution
        decomposition.communicate_ghosts(stress)

        # Apply divergence (transpose of gradient)
        force_generic.pg[...] = 0.0
        quad_weights = grad_op.quadrature_weights
        grad_op.transpose(stress, force_generic, quad_weights)

        # Compare results
        np.testing.assert_allclose(
            force_fused.p, force_generic.p, rtol=1e-10, atol=1e-10
        )


# =============================================================================
# 3D Operator Tests
# =============================================================================


class TestIsotropicStiffnessOperator3D:
    """Test suite for IsotropicStiffnessOperator3D."""

    def test_construction(self):
        """Test that the operator can be constructed."""
        grid_spacing = [0.1, 0.1, 0.1]
        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # Check that G and V matrices have correct shape
        G = op.G
        V = op.V
        assert G.shape == (24, 24)
        assert V.shape == (24, 24)

    def test_symmetry(self):
        """Test that G and V matrices are symmetric."""
        grid_spacing = [0.1, 0.2, 0.15]
        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        G = op.G
        V = op.V

        np.testing.assert_allclose(G, G.T, rtol=1e-10)
        np.testing.assert_allclose(V, V.T, rtol=1e-10)

    def test_apply_homogeneous_material(self):
        """Test apply with homogeneous material."""
        nx, ny, nz = 6, 6, 6
        grid_spacing = [1.0 / (nx - 1), 1.0 / (ny - 1), 1.0 / (nz - 1)]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # Create field collection with ghost cells on both sides
        fc = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )

        # Create fields
        displacement = fc.real_field("displacement", (3,))
        force = fc.real_field("force", (3,))

        # Create material field collection (same size as node field, with ghosts)
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))

        # Set homogeneous material properties
        lam = 1.0
        mu = 1.0
        lambda_field.p[...] = lam
        mu_field.p[...] = mu

        # Set a simple displacement field (linear in x)
        x = np.arange(nx) * grid_spacing[0]
        y = np.arange(ny) * grid_spacing[1]
        z = np.arange(nz) * grid_spacing[2]
        X, Y, Z = np.meshgrid(x, y, z, indexing="ij")
        displacement.p[0, ...] = X  # u_x = x
        displacement.p[1, ...] = 0  # u_y = 0
        displacement.p[2, ...] = 0  # u_z = 0

        # Fill ghost cells with periodic copies
        displacement.pg[...] = np.pad(
            displacement.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(
            mu_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        # Apply operator
        op.apply(displacement, lambda_field, mu_field, force)

        # Just check that it runs without error and force is finite
        assert np.all(np.isfinite(force.p))

    def test_compare_with_generic_operator(self):
        """Compare fused kernel with generic B^T C B computation."""
        nn = 5  # Number of nodes
        nel = nn  # Number of elements (periodic BC)
        grid_spacing = [0.25, 0.25, 0.25]

        # Material properties (homogeneous)
        lam = 2.0
        mu = 1.5

        # Create the fused operator
        fused_op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # Create the generic gradient operator
        grad_op = muGrid.FEMGradientOperator(3, grid_spacing)

        # Create decomposition with ghosts for periodic BC
        comm = muGrid.Communicator()
        decomposition = muGrid.CartesianDecomposition(
            comm,
            (nn, nn, nn),
            nb_subdivisions=(1, 1, 1),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            nb_sub_pts={"quad": grad_op.nb_quad_pts},
        )

        # Create material field collection with ghost cells for periodic BC
        fc_mat = muGrid.GlobalFieldCollection(
            (nel, nel, nel), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = lam
        mu_field.p[...] = mu

        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(
            mu_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        # Create displacement field
        displacement = decomposition.real_field("displacement", (3,))
        np.random.seed(42)
        displacement.p[...] = np.random.rand(*displacement.p.shape)
        decomposition.communicate_ghosts(displacement)

        # Force fields
        force_fused = decomposition.real_field("force_fused", (3,))
        force_generic = decomposition.real_field("force_generic", (3,))

        # Apply fused operator
        fused_op.apply(displacement, lambda_field, mu_field, force_fused)

        # Apply generic operator (gradient -> stress -> divergence)
        gradient = decomposition.real_field("gradient", (3, 3), "quad")
        grad_op.apply(displacement, gradient)

        # Compute stress using isotropic constitutive law
        stress = decomposition.real_field("stress", (3, 3), "quad")
        grad = gradient.s

        # Symmetric strain
        eps = np.zeros_like(stress.s)
        for i in range(3):
            for j in range(3):
                eps[i, j, ...] = 0.5 * (grad[i, j, ...] + grad[j, i, ...])

        trace = eps[0, 0, ...] + eps[1, 1, ...] + eps[2, 2, ...]

        # Stress
        for i in range(3):
            for j in range(3):
                stress.s[i, j, ...] = 2 * mu * eps[i, j, ...]
                if i == j:
                    stress.s[i, j, ...] += lam * trace

        # Fill ghost cells for periodic contribution
        decomposition.communicate_ghosts(stress)

        # Apply divergence (transpose of gradient)
        force_generic.pg[...] = 0.0
        quad_weights = grad_op.quadrature_weights
        grad_op.transpose(stress, force_generic, quad_weights)

        # Compare results
        np.testing.assert_allclose(
            force_fused.p, force_generic.p, rtol=1e-10, atol=1e-10
        )


# =============================================================================
# Apply Increment Tests
# =============================================================================


class TestIsotropicStiffnessApplyIncrement:
    """Test apply_increment functionality."""

    def test_apply_increment_2d(self):
        """Test apply_increment in 2D."""
        nx, ny = 6, 6
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Create field collections with ghosts on both sides
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        # Material field same size as node field, with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )

        displacement = fc.real_field("displacement", (2,))
        force1 = fc.real_field("force1", (2,))
        force2 = fc.real_field("force2", (2,))

        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))

        lambda_field.p[...] = 2.0
        mu_field.p[...] = 1.0
        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(mu_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        np.random.seed(42)
        displacement.pg[...] = np.random.rand(*displacement.pg.shape)

        # Apply once
        op.apply(displacement, lambda_field, mu_field, force1)

        # Apply with increment (alpha=2.0, starting from force1)
        force2.p[...] = force1.p[...]
        op.apply_increment(displacement, lambda_field, mu_field, 1.0, force2)

        # force2 should now be 2 * force1
        np.testing.assert_allclose(force2.p, 2.0 * force1.p, rtol=1e-10)

    def test_apply_increment_3d(self):
        """Test apply_increment in 3D."""
        nx, ny, nz = 5, 5, 5
        grid_spacing = [0.25, 0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # Create field collections with ghosts on both sides
        fc = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        # Material field same size as node field, with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )

        displacement = fc.real_field("displacement", (3,))
        force1 = fc.real_field("force1", (3,))
        force2 = fc.real_field("force2", (3,))

        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))

        lambda_field.p[...] = 2.0
        mu_field.p[...] = 1.0
        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(
            mu_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        np.random.seed(42)
        displacement.pg[...] = np.random.rand(*displacement.pg.shape)

        # Apply once
        op.apply(displacement, lambda_field, mu_field, force1)

        # Apply with increment (alpha=0.5, starting from zero)
        force2.p[...] = 0.0
        op.apply_increment(displacement, lambda_field, mu_field, 0.5, force2)

        # force2 should be 0.5 * force1
        np.testing.assert_allclose(force2.p, 0.5 * force1.p, rtol=1e-10)


# =============================================================================
# Smoke Tests (Parametrized on Device)
# =============================================================================


@pytest.mark.parametrize("device", get_test_devices())
class TestIsotropicStiffnessOperatorApply_smoke:
    """Smoke tests for IsotropicStiffnessOperator apply on different devices."""

    def test_2d_apply_smoke(self, device):
        """Test 2D operator apply smoke test."""
        skip_if_gpu_unavailable(device)

        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Create device field collections
        device_obj = create_device(device)
        fc_kwargs = {"device": device_obj} if device_obj else {}

        fc = muGrid.GlobalFieldCollection(
            (nx, ny),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            **fc_kwargs,
        )
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            **fc_kwargs,
        )

        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))

        # Initialize
        displacement.set_zero()
        lambda_field.pg[...] = 2.0
        mu_field.pg[...] = 1.0

        # Apply should not raise
        op.apply(displacement, lambda_field, mu_field, force)

        # Check that force is finite
        if device == "gpu":
            assert force.is_on_gpu
            force_data = cp.asnumpy(force.p)
        else:
            force_data = force.p
        assert np.all(np.isfinite(force_data))

    def test_3d_apply_smoke(self, device):
        """Test 3D operator apply smoke test."""
        skip_if_gpu_unavailable(device)

        nx, ny, nz = 6, 6, 6
        grid_spacing = [0.25, 0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # Create device field collections
        device_obj = create_device(device)
        fc_kwargs = {"device": device_obj} if device_obj else {}

        fc = muGrid.GlobalFieldCollection(
            (nx, ny, nz),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            **fc_kwargs,
        )
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny, nz),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            **fc_kwargs,
        )

        displacement = fc.real_field("displacement", (3,))
        force = fc.real_field("force", (3,))
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))

        # Initialize
        displacement.set_zero()
        lambda_field.pg[...] = 2.0
        mu_field.pg[...] = 1.0

        # Apply should not raise
        op.apply(displacement, lambda_field, mu_field, force)

        # Check that force is finite
        if device == "gpu":
            assert force.is_on_gpu
            force_data = cp.asnumpy(force.p)
        else:
            force_data = force.p
        assert np.all(np.isfinite(force_data))


# =============================================================================
# GPU vs CPU Consistency Tests (Dual-device comparison)
# =============================================================================


@pytest.mark.skipif(not GPU_AVAILABLE, reason="GPU backend not available")
class TestIsotropicStiffnessOperatorGPUCorrectness:
    """Test GPU correctness by comparing with CPU."""

    def setup_method(self, method):
        """Skip tests if no GPU is available at runtime."""
        if not muGrid.is_gpu_available():
            pytest.skip("No GPU device available at runtime")
        if not HAS_CUPY:
            pytest.skip("CuPy not available for GPU tests")

    def test_gpu_cpu_consistency_2d(self):
        """Test that GPU and CPU give same results in 2D."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Both CPU and GPU use same ghost configuration for comparison
        fc_cpu = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        fc_mat_cpu = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )

        disp_cpu = fc_cpu.real_field("displacement", (2,))
        force_cpu = fc_cpu.real_field("force", (2,))
        lam_cpu = fc_mat_cpu.real_field("lambda", (1,))
        mu_cpu = fc_mat_cpu.real_field("mu", (1,))

        fc_gpu = muGrid.GlobalFieldCollection(
            (nx, ny),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )
        fc_mat_gpu = muGrid.GlobalFieldCollection(
            (nx, ny),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )

        disp_gpu = fc_gpu.real_field("displacement", (2,))
        force_gpu = fc_gpu.real_field("force", (2,))
        lam_gpu = fc_mat_gpu.real_field("lambda", (1,))
        mu_gpu = fc_mat_gpu.real_field("mu", (1,))

        # Initialize with same data
        np.random.seed(42)
        test_disp = np.random.rand(*disp_cpu.pg.shape)
        test_lam = 2.0 + np.random.rand(*lam_cpu.p.shape)
        test_mu = 1.0 + np.random.rand(*mu_cpu.p.shape)

        disp_cpu.pg[...] = test_disp
        lam_cpu.p[...] = test_lam
        mu_cpu.p[...] = test_mu
        # Fill ghost cells with periodic copies
        lam_cpu.pg[...] = np.pad(lam_cpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")
        mu_cpu.pg[...] = np.pad(mu_cpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        disp_gpu.pg[...] = cp.asarray(test_disp)
        lam_gpu.p[...] = cp.asarray(test_lam)
        mu_gpu.p[...] = cp.asarray(test_mu)
        # Fill ghost cells with periodic copies on GPU
        lam_gpu.pg[...] = cp.pad(lam_gpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")
        mu_gpu.pg[...] = cp.pad(mu_gpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        # Apply
        op.apply(disp_cpu, lam_cpu, mu_cpu, force_cpu)
        op.apply(disp_gpu, lam_gpu, mu_gpu, force_gpu)

        # Compare
        force_gpu_np = cp.asnumpy(force_gpu.p)
        np.testing.assert_allclose(force_gpu_np, force_cpu.p, rtol=1e-10)

    def test_gpu_cpu_consistency_3d(self):
        """Test that GPU and CPU give same results in 3D."""
        nx, ny, nz = 6, 6, 6
        grid_spacing = [0.25, 0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # Both CPU and GPU use same ghost configuration for comparison
        fc_cpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        fc_mat_cpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )

        disp_cpu = fc_cpu.real_field("displacement", (3,))
        force_cpu = fc_cpu.real_field("force", (3,))
        lam_cpu = fc_mat_cpu.real_field("lambda", (1,))
        mu_cpu = fc_mat_cpu.real_field("mu", (1,))

        fc_gpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            device=muGrid.Device.gpu(),
        )
        fc_mat_gpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            device=muGrid.Device.gpu(),
        )

        disp_gpu = fc_gpu.real_field("displacement", (3,))
        force_gpu = fc_gpu.real_field("force", (3,))
        lam_gpu = fc_mat_gpu.real_field("lambda", (1,))
        mu_gpu = fc_mat_gpu.real_field("mu", (1,))

        # Initialize with same data
        np.random.seed(42)
        test_disp = np.random.rand(*disp_cpu.pg.shape)
        test_lam = 2.0 + np.random.rand(*lam_cpu.p.shape)
        test_mu = 1.0 + np.random.rand(*mu_cpu.p.shape)

        disp_cpu.pg[...] = test_disp
        lam_cpu.p[...] = test_lam
        mu_cpu.p[...] = test_mu
        # Fill ghost cells with periodic copies
        lam_cpu.pg[...] = np.pad(
            lam_cpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_cpu.pg[...] = np.pad(
            mu_cpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        disp_gpu.pg[...] = cp.asarray(test_disp)
        lam_gpu.p[...] = cp.asarray(test_lam)
        mu_gpu.p[...] = cp.asarray(test_mu)
        # Fill ghost cells with periodic copies on GPU
        lam_gpu.pg[...] = cp.pad(
            lam_gpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_gpu.pg[...] = cp.pad(
            mu_gpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        # Apply
        op.apply(disp_cpu, lam_cpu, mu_cpu, force_cpu)
        op.apply(disp_gpu, lam_gpu, mu_gpu, force_gpu)

        # Compare
        force_gpu_np = cp.asnumpy(force_gpu.p)
        np.testing.assert_allclose(force_gpu_np, force_cpu.p, rtol=1e-10)


# =============================================================================
# Unit Impulse Tests (systematic comparison with generic operator)
# =============================================================================


class TestUnitImpulse2D:
    """Test fused operator matches generic operator for unit impulses in 2D."""

    def _generic_apply(self, decomposition, displacement, force, lam, mu, grad_op):
        """Apply generic B^T C B operator using FEMGradientOperator."""
        gradient = decomposition.real_field("gradient", (2, 2), "quad")
        stress = decomposition.real_field("stress", (2, 2), "quad")

        grad_op.apply(displacement, gradient)
        grad = gradient.s

        # Symmetric strain
        eps_xx = grad[0, 0, ...]
        eps_yy = grad[1, 1, ...]
        eps_xy = 0.5 * (grad[0, 1, ...] + grad[1, 0, ...])
        trace = eps_xx + eps_yy

        # Stress (isotropic constitutive law)
        stress.s[0, 0, ...] = lam * trace + 2 * mu * eps_xx
        stress.s[1, 1, ...] = lam * trace + 2 * mu * eps_yy
        stress.s[0, 1, ...] = 2 * mu * eps_xy
        stress.s[1, 0, ...] = 2 * mu * eps_xy

        # Fill stress ghost cells for periodic contributions
        decomposition.communicate_ghosts(stress)

        # Divergence
        force.pg[...] = 0.0
        quad_weights = grad_op.quadrature_weights
        grad_op.transpose(stress, force, quad_weights)

    def _convolution_apply(
        self, decomposition, displacement, force, lam, mu, conv_op, quad_weights
    ):
        """Apply B^T C B operator using ConvolutionOperator.

        This uses a ConvolutionOperator created from FEMGradientOperator.coefficients
        to verify that both gradient operators produce identical results.

        Note: ConvolutionOperator.transpose doesn't apply quadrature weights internally,
        so we pre-multiply the stress by weights before calling transpose.
        """
        gradient = decomposition.real_field("gradient_conv", (2, 2), "quad")
        stress = decomposition.real_field("stress_conv", (2, 2), "quad")

        conv_op.apply(displacement, gradient)
        grad = gradient.s

        # Symmetric strain (same as _generic_apply)
        eps_xx = grad[0, 0, ...]
        eps_yy = grad[1, 1, ...]
        eps_xy = 0.5 * (grad[0, 1, ...] + grad[1, 0, ...])
        trace = eps_xx + eps_yy

        # Stress (isotropic constitutive law)
        stress.s[0, 0, ...] = lam * trace + 2 * mu * eps_xx
        stress.s[1, 1, ...] = lam * trace + 2 * mu * eps_yy
        stress.s[0, 1, ...] = 2 * mu * eps_xy
        stress.s[1, 0, ...] = 2 * mu * eps_xy

        # Pre-multiply stress by quadrature weights (ConvolutionOperator.transpose
        # doesn't apply weights internally like FEMGradientOperator does)
        for q in range(len(quad_weights)):
            stress.s[:, :, q, ...] *= quad_weights[q]

        # Fill stress ghost cells for periodic contributions
        decomposition.communicate_ghosts(stress)

        # Divergence using transpose of ConvolutionOperator (without weights)
        force.pg[...] = 0.0
        conv_op.transpose(stress, force, [])

    def test_unit_impulse_non_periodic(self):
        """Test unit impulse response for non-periodic BC.

        With node-based material field indexing, non-periodic vs periodic is
        distinguished only by how ghost cells are filled (not by kernel logic).
        This test uses zero-filled ghosts to simulate non-periodic BC.

        Compares three approaches:
        1. Fused IsotropicStiffnessOperator2D
        2. Generic FEMGradientOperator
        3. ConvolutionOperator with FEMGradientOperator.coefficients

        Also outputs timing information for the generic vs convolution approaches.
        """
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        lam = 2.0
        mu = 1.5

        fused_op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)
        grad_op = muGrid.FEMGradientOperator(2, grid_spacing)

        # Create ConvolutionOperator from FEMGradientOperator coefficients
        conv_op = muGrid.GenericLinearOperator([0, 0], grad_op.coefficients)
        quad_weights = grad_op.quadrature_weights

        # Use CartesianDecomposition for ghost handling
        comm = muGrid.Communicator()
        decomposition = muGrid.CartesianDecomposition(
            comm,
            (nx, ny),
            nb_subdivisions=(1, 1),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            nb_sub_pts={"quad": grad_op.nb_quad_pts},
        )

        # Material field same size as node field, with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )

        displacement = decomposition.real_field("displacement", (2,))
        force_fused = decomposition.real_field("force_fused", (2,))
        force_generic = decomposition.real_field("force_generic", (2,))
        force_convolution = decomposition.real_field("force_convolution", (2,))

        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = lam
        mu_field.p[...] = mu
        # Fill ghost cells with periodic copies (same as periodic BC for material)
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(mu_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        timer = Timer()

        # Test unit impulses at a few interior nodes
        test_positions = [(2, 3), (4, 4), (5, 2)]
        for ix, iy in test_positions:
            for d in range(2):  # Test both displacement directions
                displacement.p[...] = 0.0
                displacement.p[d, ix, iy] = 1.0

                # communicate_ghosts for ghost handling
                decomposition.communicate_ghosts(displacement)

                force_fused.p[...] = 0.0
                force_generic.p[...] = 0.0
                force_convolution.p[...] = 0.0

                with timer("fused"):
                    fused_op.apply(displacement, lambda_field, mu_field, force_fused)

                with timer("FEMGradientOperator"):
                    self._generic_apply(
                        decomposition, displacement, force_generic, lam, mu, grad_op
                    )

                with timer("ConvolutionOperator"):
                    self._convolution_apply(
                        decomposition,
                        displacement,
                        force_convolution,
                        lam,
                        mu,
                        conv_op,
                        quad_weights,
                    )

                # Compare fused vs generic FEMGradientOperator
                # With node-based indexing, boundary forces are computed too
                np.testing.assert_allclose(
                    force_fused.p,
                    force_generic.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"Fused vs generic mismatch at node ({ix}, {iy}), "
                    f"direction {d}",
                )

                # Compare generic FEMGradientOperator vs ConvolutionOperator
                np.testing.assert_allclose(
                    force_generic.p,
                    force_convolution.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"Generic vs convolution mismatch at node ({ix}, {iy}), "
                    f"direction {d}",
                )

        timer.print_summary(title=f"2D Non-periodic ({nx}x{ny} grid)")

    def test_unit_impulse_periodic(self):
        """Test unit impulse response for periodic BC (N elements for N nodes).

        Compares three approaches:
        1. Fused IsotropicStiffnessOperator2D
        2. Generic FEMGradientOperator
        3. ConvolutionOperator with FEMGradientOperator.coefficients

        Also outputs timing information for the generic vs convolution approaches.
        """
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        lam = 2.0
        mu = 1.5

        fused_op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)
        grad_op = muGrid.FEMGradientOperator(2, grid_spacing)

        # Create ConvolutionOperator from FEMGradientOperator coefficients
        conv_op = muGrid.GenericLinearOperator([0, 0], grad_op.coefficients)
        quad_weights = grad_op.quadrature_weights

        # Use CartesianDecomposition to handle ghost communication for periodic BC
        comm = muGrid.Communicator()
        decomposition = muGrid.CartesianDecomposition(
            comm,
            (nx, ny),
            nb_subdivisions=(1, 1),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            nb_sub_pts={"quad": grad_op.nb_quad_pts},
        )

        # Periodic: material field has N elements (same as nodes) with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )

        displacement = decomposition.real_field("displacement", (2,))
        force_fused = decomposition.real_field("force_fused", (2,))
        force_generic = decomposition.real_field("force_generic", (2,))
        force_convolution = decomposition.real_field("force_convolution", (2,))

        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = lam
        mu_field.p[...] = mu

        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(mu_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        timer = Timer()

        # Test unit impulses at various nodes including boundaries
        # For periodic BC, boundary nodes should also match
        test_positions = [(0, 0), (7, 7), (0, 4), (4, 0), (3, 5)]
        for ix, iy in test_positions:
            for d in range(2):
                # Set displacement and use communicate_ghosts for periodic filling
                displacement.p[...] = 0.0
                displacement.p[d, ix, iy] = 1.0

                # communicate_ghosts fills ghost cells with periodic copies
                decomposition.communicate_ghosts(displacement)

                force_fused.p[...] = 0.0
                force_generic.p[...] = 0.0
                force_convolution.p[...] = 0.0

                with timer("fused"):
                    fused_op.apply(displacement, lambda_field, mu_field, force_fused)

                with timer("FEMGradientOperator"):
                    self._generic_apply(
                        decomposition, displacement, force_generic, lam, mu, grad_op
                    )

                with timer("ConvolutionOperator"):
                    self._convolution_apply(
                        decomposition,
                        displacement,
                        force_convolution,
                        lam,
                        mu,
                        conv_op,
                        quad_weights,
                    )

                # Compare fused vs generic FEMGradientOperator
                np.testing.assert_allclose(
                    force_fused.p,
                    force_generic.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"Fused vs generic mismatch at node ({ix}, {iy}), "
                    f"direction {d}",
                )

                # Compare generic FEMGradientOperator vs ConvolutionOperator
                np.testing.assert_allclose(
                    force_generic.p,
                    force_convolution.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"Generic vs convolution mismatch at node ({ix}, {iy}), "
                    f"direction {d}",
                )

        timer.print_summary(title=f"2D Periodic ({nx}x{ny} grid)")


class TestUnitImpulse3D:
    """Test fused operator matches generic operator for unit impulses in 3D."""

    def _generic_apply(self, decomposition, displacement, force, lam, mu, grad_op):
        """Apply generic B^T C B operator using FEMGradientOperator."""
        gradient = decomposition.real_field("gradient", (3, 3), "quad")
        stress = decomposition.real_field("stress", (3, 3), "quad")

        grad_op.apply(displacement, gradient)
        grad = gradient.s

        # Symmetric strain
        eps = np.zeros_like(stress.s)
        for i in range(3):
            for j in range(3):
                eps[i, j, ...] = 0.5 * (grad[i, j, ...] + grad[j, i, ...])

        trace = eps[0, 0, ...] + eps[1, 1, ...] + eps[2, 2, ...]

        # Stress
        for i in range(3):
            for j in range(3):
                stress.s[i, j, ...] = 2 * mu * eps[i, j, ...]
                if i == j:
                    stress.s[i, j, ...] += lam * trace

        # Fill stress ghost cells for periodic contributions
        decomposition.communicate_ghosts(stress)

        # Divergence
        force.pg[...] = 0.0
        quad_weights = grad_op.quadrature_weights
        grad_op.transpose(stress, force, quad_weights)

    def _convolution_apply(
        self, decomposition, displacement, force, lam, mu, conv_op, quad_weights
    ):
        """Apply B^T C B operator using ConvolutionOperator.

        This uses a ConvolutionOperator created from FEMGradientOperator.coefficients
        to verify that both gradient operators produce identical results.

        Note: ConvolutionOperator.transpose doesn't apply quadrature weights internally,
        so we pre-multiply the stress by weights before calling transpose.
        """
        gradient = decomposition.real_field("gradient_conv", (3, 3), "quad")
        stress = decomposition.real_field("stress_conv", (3, 3), "quad")

        conv_op.apply(displacement, gradient)
        grad = gradient.s

        # Symmetric strain (same as _generic_apply)
        eps = np.zeros_like(stress.s)
        for i in range(3):
            for j in range(3):
                eps[i, j, ...] = 0.5 * (grad[i, j, ...] + grad[j, i, ...])

        trace = eps[0, 0, ...] + eps[1, 1, ...] + eps[2, 2, ...]

        # Stress (isotropic constitutive law)
        for i in range(3):
            for j in range(3):
                stress.s[i, j, ...] = 2 * mu * eps[i, j, ...]
                if i == j:
                    stress.s[i, j, ...] += lam * trace

        # Pre-multiply stress by quadrature weights (ConvolutionOperator.transpose
        # doesn't apply weights internally like FEMGradientOperator does)
        for q in range(len(quad_weights)):
            stress.s[:, :, q, ...] *= quad_weights[q]

        # Fill stress ghost cells for periodic contributions
        decomposition.communicate_ghosts(stress)

        # Divergence using transpose of ConvolutionOperator (without weights)
        force.pg[...] = 0.0
        conv_op.transpose(stress, force, [])

    def test_unit_impulse_non_periodic(self):
        """Test unit impulse response for non-periodic BC in 3D.

        With node-based material field indexing, non-periodic vs periodic is
        distinguished only by how ghost cells are filled (not by kernel logic).

        Compares three approaches:
        1. Fused IsotropicStiffnessOperator3D
        2. Generic FEMGradientOperator
        3. ConvolutionOperator with FEMGradientOperator.coefficients

        Also outputs timing information for the generic vs convolution approaches.
        """
        nx, ny, nz = 6, 6, 6
        grid_spacing = [0.25, 0.25, 0.25]

        lam = 2.0
        mu = 1.5

        fused_op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)
        grad_op = muGrid.FEMGradientOperator(3, grid_spacing)

        # Create ConvolutionOperator from FEMGradientOperator coefficients
        conv_op = muGrid.GenericLinearOperator([0, 0, 0], grad_op.coefficients)
        quad_weights = grad_op.quadrature_weights

        # Use CartesianDecomposition for ghost handling
        comm = muGrid.Communicator()
        decomposition = muGrid.CartesianDecomposition(
            comm,
            (nx, ny, nz),
            nb_subdivisions=(1, 1, 1),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            nb_sub_pts={"quad": grad_op.nb_quad_pts},
        )

        # Material field same size as node field, with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )

        displacement = decomposition.real_field("displacement", (3,))
        force_fused = decomposition.real_field("force_fused", (3,))
        force_generic = decomposition.real_field("force_generic", (3,))
        force_convolution = decomposition.real_field("force_convolution", (3,))

        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = lam
        mu_field.p[...] = mu
        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(
            mu_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        timer = Timer()

        # Test a few interior nodes
        test_positions = [(2, 2, 2), (3, 2, 4)]
        for ix, iy, iz in test_positions:
            for d in range(3):
                displacement.p[...] = 0.0
                displacement.p[d, ix, iy, iz] = 1.0

                decomposition.communicate_ghosts(displacement)

                force_fused.p[...] = 0.0
                force_generic.p[...] = 0.0
                force_convolution.p[...] = 0.0

                with timer("fused"):
                    fused_op.apply(displacement, lambda_field, mu_field, force_fused)

                with timer("FEMGradientOperator"):
                    self._generic_apply(
                        decomposition, displacement, force_generic, lam, mu, grad_op
                    )

                with timer("ConvolutionOperator"):
                    self._convolution_apply(
                        decomposition,
                        displacement,
                        force_convolution,
                        lam,
                        mu,
                        conv_op,
                        quad_weights,
                    )

                # Compare fused vs generic FEMGradientOperator
                # With node-based indexing, boundary forces are computed too
                np.testing.assert_allclose(
                    force_fused.p,
                    force_generic.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"Fused vs generic mismatch at node ({ix}, {iy}, {iz}), "
                    f"direction {d}",
                )

                # Compare generic FEMGradientOperator vs ConvolutionOperator
                np.testing.assert_allclose(
                    force_generic.p,
                    force_convolution.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg="Generic vs convolution mismatch at node "
                    f"({ix}, {iy}, {iz}), "
                    f"direction {d}",
                )

        timer.print_summary(title=f"3D Non-periodic ({nx}x{ny}x{nz} grid)")

    def test_unit_impulse_periodic(self):
        """Test unit impulse response for periodic BC in 3D.

        Compares three approaches:
        1. Fused IsotropicStiffnessOperator3D
        2. Generic FEMGradientOperator
        3. ConvolutionOperator with FEMGradientOperator.coefficients
        """
        nx, ny, nz = 5, 5, 5
        grid_spacing = [0.25, 0.25, 0.25]

        lam = 2.0
        mu = 1.5

        fused_op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)
        grad_op = muGrid.FEMGradientOperator(3, grid_spacing)

        # Create ConvolutionOperator from FEMGradientOperator coefficients
        conv_op = muGrid.GenericLinearOperator([0, 0, 0], grad_op.coefficients)
        quad_weights = grad_op.quadrature_weights

        # Use CartesianDecomposition to handle ghost communication for periodic BC
        comm = muGrid.Communicator()
        decomposition = muGrid.CartesianDecomposition(
            comm,
            (nx, ny, nz),
            nb_subdivisions=(1, 1, 1),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            nb_sub_pts={"quad": grad_op.nb_quad_pts},
        )

        # Periodic: material field with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )

        displacement = decomposition.real_field("displacement", (3,))
        force_fused = decomposition.real_field("force_fused", (3,))
        force_generic = decomposition.real_field("force_generic", (3,))
        force_convolution = decomposition.real_field("force_convolution", (3,))

        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = lam
        mu_field.p[...] = mu

        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(
            mu_field.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        timer = Timer()

        # Test boundary and interior nodes
        test_positions = [(0, 0, 0), (4, 4, 4), (2, 3, 1)]
        for ix, iy, iz in test_positions:
            for d in range(3):
                # Set displacement and use communicate_ghosts for periodic filling
                displacement.p[...] = 0.0
                displacement.p[d, ix, iy, iz] = 1.0

                # communicate_ghosts fills ghost cells with periodic copies
                decomposition.communicate_ghosts(displacement)

                force_fused.p[...] = 0.0
                force_generic.p[...] = 0.0
                force_convolution.p[...] = 0.0

                with timer("fused"):
                    fused_op.apply(displacement, lambda_field, mu_field, force_fused)

                with timer("FEMGradientOperator"):
                    self._generic_apply(
                        decomposition, displacement, force_generic, lam, mu, grad_op
                    )

                with timer("ConvolutionOperator"):
                    self._convolution_apply(
                        decomposition,
                        displacement,
                        force_convolution,
                        lam,
                        mu,
                        conv_op,
                        quad_weights,
                    )

                # Compare fused vs generic FEMGradientOperator
                np.testing.assert_allclose(
                    force_fused.p,
                    force_generic.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"Fused vs generic mismatch at node ({ix}, {iy}, {iz}), "
                    f"direction {d}",
                )

                # Compare generic FEMGradientOperator vs ConvolutionOperator
                np.testing.assert_allclose(
                    force_generic.p,
                    force_convolution.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg="Generic vs convolution mismatch at node "
                    f"({ix}, {iy}, {iz}), "
                    f"direction {d}",
                )

        timer.print_summary(title=f"3D Periodic ({nx}x{ny}x{nz} grid)")


# =============================================================================
# Validation Guard Tests
# =============================================================================


class TestValidationGuard2D:
    """Test that validation guards raise appropriate exceptions in 2D."""

    def test_wrong_material_field_size(self):
        """Test error when material field has wrong size."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Displacement field with proper ghosts
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))

        # Material field with wrong size (not same as node field)
        fc_mat_wrong = muGrid.GlobalFieldCollection(
            (nx - 2, ny - 2), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        lambda_field = fc_mat_wrong.real_field("lambda", (1,))
        mu_field = fc_mat_wrong.real_field("mu", (1,))
        lambda_field.pg[...] = 1.0
        mu_field.pg[...] = 1.0

        with pytest.raises(RuntimeError) as exc_info:
            op.apply(displacement, lambda_field, mu_field, force)
        err_msg = str(exc_info.value).lower()
        assert "material field" in err_msg or "dimensions" in err_msg

    def test_missing_left_ghosts_node(self):
        """Test error when left ghost cells are missing for node field."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Only right ghosts for node field - this is invalid
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(0, 0), nb_ghosts_right=(1, 1)
        )
        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))

        # Proper material field with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.pg[...] = 1.0
        mu_field.pg[...] = 1.0

        with pytest.raises(RuntimeError) as exc_info:
            op.apply(displacement, lambda_field, mu_field, force)
        err_msg = str(exc_info.value).lower()
        assert "ghost" in err_msg or "ghosts" in err_msg

    def test_missing_right_ghosts(self):
        """Test error when right ghost cells are missing."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # No right ghosts for node field
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(0, 0)
        )
        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))

        # Proper material field with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.pg[...] = 1.0
        mu_field.pg[...] = 1.0

        with pytest.raises(RuntimeError) as exc_info:
            op.apply(displacement, lambda_field, mu_field, force)
        assert "ghost" in str(exc_info.value).lower()

    def test_missing_material_ghosts(self):
        """Test error when ghost cells are missing for material field."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Proper node field with ghost cells
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))

        # Material field without ghost cells
        fc_mat = muGrid.GlobalFieldCollection((nx, ny))
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.s[...] = 1.0
        mu_field.s[...] = 1.0

        with pytest.raises(RuntimeError) as exc_info:
            op.apply(displacement, lambda_field, mu_field, force)
        err_msg = str(exc_info.value).lower()
        assert "ghost" in err_msg or "ghosts" in err_msg

    def test_valid_config(self):
        """Test that valid configuration does not raise."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # Node field with ghost cells
        fc = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        displacement = fc.real_field("displacement", (2,))
        force = fc.real_field("force", (2,))

        # Material field same size as node field, with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.p[...] = 1.0
        mu_field.p[...] = 1.0

        # Fill ghost cells with periodic copies
        lambda_field.pg[...] = np.pad(
            lambda_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap"
        )
        mu_field.pg[...] = np.pad(mu_field.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        # Should not raise
        op.apply(displacement, lambda_field, mu_field, force)


class TestValidationGuard3D:
    """Test that validation guards raise appropriate exceptions in 3D."""

    def test_wrong_material_field_size(self):
        """Test error when material field has wrong size in 3D."""
        nx, ny, nz = 6, 6, 6
        grid_spacing = [0.25, 0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        fc = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        displacement = fc.real_field("displacement", (3,))
        force = fc.real_field("force", (3,))

        # Wrong size (not same as node field)
        fc_mat_wrong = muGrid.GlobalFieldCollection(
            (nx - 2, ny - 2, nz - 2),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
        )
        lambda_field = fc_mat_wrong.real_field("lambda", (1,))
        mu_field = fc_mat_wrong.real_field("mu", (1,))
        lambda_field.pg[...] = 1.0
        mu_field.pg[...] = 1.0

        with pytest.raises(RuntimeError) as exc_info:
            op.apply(displacement, lambda_field, mu_field, force)
        err_msg = str(exc_info.value).lower()
        assert "material field" in err_msg or "dimensions" in err_msg

    def test_missing_ghosts(self):
        """Test error when ghost cells are missing in 3D."""
        nx, ny, nz = 6, 6, 6
        grid_spacing = [0.25, 0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # No ghosts at all for node field
        fc = muGrid.GlobalFieldCollection((nx, ny, nz))
        displacement = fc.real_field("displacement", (3,))
        force = fc.real_field("force", (3,))

        # Proper material field with ghost cells
        fc_mat = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        lambda_field = fc_mat.real_field("lambda", (1,))
        mu_field = fc_mat.real_field("mu", (1,))
        lambda_field.pg[...] = 1.0
        mu_field.pg[...] = 1.0

        with pytest.raises(RuntimeError) as exc_info:
            op.apply(displacement, lambda_field, mu_field, force)
        assert "ghost" in str(exc_info.value).lower()


# =============================================================================
# GPU Unit Impulse Tests (Dual-device comparison)
# =============================================================================


@pytest.mark.skipif(not GPU_AVAILABLE, reason="GPU backend not available")
class TestGPUUnitImpulse:
    """Test GPU unit impulse response matches CPU."""

    def setup_method(self, method):
        """Skip tests if no GPU is available at runtime."""
        if not muGrid.is_gpu_available():
            pytest.skip("No GPU device available at runtime")
        if not HAS_CUPY:
            pytest.skip("CuPy not available for GPU tests")

    def test_gpu_unit_impulse_2d(self):
        """Test GPU unit impulse response in 2D."""
        nx, ny = 8, 8
        grid_spacing = [0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator2D(grid_spacing)

        # GPU kernels require ghosts on both sides
        fc_cpu = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )
        # Material field same size as node field, with ghost cells
        fc_mat_cpu = muGrid.GlobalFieldCollection(
            (nx, ny), nb_ghosts_left=(1, 1), nb_ghosts_right=(1, 1)
        )

        disp_cpu = fc_cpu.real_field("displacement", (2,))
        force_cpu = fc_cpu.real_field("force", (2,))
        lam_cpu = fc_mat_cpu.real_field("lambda", (1,))
        mu_cpu = fc_mat_cpu.real_field("mu", (1,))

        lam_cpu.p[...] = 2.0
        mu_cpu.p[...] = 1.5
        # Fill ghost cells with periodic copies
        lam_cpu.pg[...] = np.pad(lam_cpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")
        mu_cpu.pg[...] = np.pad(mu_cpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        # GPU setup
        fc_gpu = muGrid.GlobalFieldCollection(
            (nx, ny),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )
        fc_mat_gpu = muGrid.GlobalFieldCollection(
            (nx, ny),
            nb_ghosts_left=(1, 1),
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )

        disp_gpu = fc_gpu.real_field("displacement", (2,))
        force_gpu = fc_gpu.real_field("force", (2,))
        lam_gpu = fc_mat_gpu.real_field("lambda", (1,))
        mu_gpu = fc_mat_gpu.real_field("mu", (1,))

        lam_gpu.p[...] = 2.0
        mu_gpu.p[...] = 1.5
        # Fill ghost cells with periodic copies on GPU
        lam_gpu.pg[...] = cp.pad(lam_gpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")
        mu_gpu.pg[...] = cp.pad(mu_gpu.p, ((0, 0), (1, 1), (1, 1)), mode="wrap")

        # Test a few unit impulses
        test_positions = [(2, 2), (3, 4), (5, 5)]
        for ix, iy in test_positions:
            for d in range(2):
                # Set up CPU and GPU data independently (same logical values)
                disp_cpu.pg[...] = 0.0
                disp_cpu.pg[d, ix + 1, iy + 1] = 1.0

                disp_gpu.pg[...] = 0.0
                disp_gpu.pg[d, ix + 1, iy + 1] = 1.0

                force_cpu.p[...] = 0.0
                force_gpu.p[...] = 0.0

                op.apply(disp_cpu, lam_cpu, mu_cpu, force_cpu)
                op.apply(disp_gpu, lam_gpu, mu_gpu, force_gpu)

                force_gpu_np = cp.asnumpy(force_gpu.p)
                np.testing.assert_allclose(
                    force_gpu_np,
                    force_cpu.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"GPU mismatch at node ({ix}, {iy}), direction {d}",
                )

    def test_gpu_unit_impulse_3d(self):
        """Test GPU unit impulse response in 3D."""
        nx, ny, nz = 6, 6, 6
        grid_spacing = [0.25, 0.25, 0.25]

        op = muGrid.IsotropicStiffnessOperator3D(grid_spacing)

        # GPU kernels require ghosts on both sides
        fc_cpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )
        # Material field same size as node field, with ghost cells
        fc_mat_cpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz), nb_ghosts_left=(1, 1, 1), nb_ghosts_right=(1, 1, 1)
        )

        disp_cpu = fc_cpu.real_field("displacement", (3,))
        force_cpu = fc_cpu.real_field("force", (3,))
        lam_cpu = fc_mat_cpu.real_field("lambda", (1,))
        mu_cpu = fc_mat_cpu.real_field("mu", (1,))

        lam_cpu.p[...] = 2.0
        mu_cpu.p[...] = 1.5
        # Fill ghost cells with periodic copies
        lam_cpu.pg[...] = np.pad(
            lam_cpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_cpu.pg[...] = np.pad(
            mu_cpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        # GPU setup
        fc_gpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            device=muGrid.Device.gpu(),
        )
        fc_mat_gpu = muGrid.GlobalFieldCollection(
            (nx, ny, nz),
            nb_ghosts_left=(1, 1, 1),
            nb_ghosts_right=(1, 1, 1),
            device=muGrid.Device.gpu(),
        )

        disp_gpu = fc_gpu.real_field("displacement", (3,))
        force_gpu = fc_gpu.real_field("force", (3,))
        lam_gpu = fc_mat_gpu.real_field("lambda", (1,))
        mu_gpu = fc_mat_gpu.real_field("mu", (1,))

        lam_gpu.p[...] = 2.0
        mu_gpu.p[...] = 1.5
        # Fill ghost cells with periodic copies on GPU
        lam_gpu.pg[...] = cp.pad(
            lam_gpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )
        mu_gpu.pg[...] = cp.pad(
            mu_gpu.p, ((0, 0), (1, 1), (1, 1), (1, 1)), mode="wrap"
        )

        # Test a few unit impulses
        test_positions = [(2, 2, 2), (3, 2, 3), (4, 3, 2)]
        for ix, iy, iz in test_positions:
            for d in range(3):
                # Set up CPU and GPU data independently (same logical values)
                disp_cpu.pg[...] = 0.0
                disp_cpu.pg[d, ix + 1, iy + 1, iz + 1] = 1.0

                disp_gpu.pg[...] = 0.0
                disp_gpu.pg[d, ix + 1, iy + 1, iz + 1] = 1.0

                force_cpu.p[...] = 0.0
                force_gpu.p[...] = 0.0

                op.apply(disp_cpu, lam_cpu, mu_cpu, force_cpu)
                op.apply(disp_gpu, lam_gpu, mu_gpu, force_gpu)

                force_gpu_np = cp.asnumpy(force_gpu.p)
                np.testing.assert_allclose(
                    force_gpu_np,
                    force_cpu.p,
                    rtol=1e-10,
                    atol=1e-14,
                    err_msg=f"GPU mismatch at node ({ix}, {iy}, {iz}), direction {d}",
                )


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import numpy as np
import muGrid

# Two dimensional grid
nb_grid_pts = (11, 12, 13)
fc = muGrid.GlobalFieldCollection(nb_grid_pts, nb_sub_pts={"element": 5})

# Get a tensor-field (for example to represent the strain)
strain = fc.real_field("strain", (3, 3), "element")

# Fill the field with random numbers
strain.s[...] = np.random.rand(*((3, 3, 5) + nb_grid_pts))

# Note that the convenience accessor `p` yields the same field but a different shape
strain.p[...] = np.random.rand(*((3, 3 * 5) + nb_grid_pts))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   __init__.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   10 Oct 2019

@brief  Main entry point for muGrid Python module

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None

# Import the C++ extension module
# Try relative import first (for installed wheels where _muGrid.so is in the
# package). Fall back to absolute import (for development where _muGrid.so is
# in build directory)
try:
    from . import _muGrid
except ImportError:
    import _muGrid

has_mpi = _muGrid.Communicator.has_mpi
if has_mpi and MPI is None:
    raise RuntimeError("MPI support is enabled for muGrid but mpi4py is not available.")

# Feature flags for compile-time configuration
has_cuda = _muGrid.has_cuda
has_rocm = _muGrid.has_rocm
has_gpu = _muGrid.has_gpu
has_netcdf = _muGrid.has_netcdf
host_arch = _muGrid.host_arch
device_arch = _muGrid.device_arch

# Device abstraction
Device = _muGrid.Device
DeviceType = _muGrid.DeviceType


# Cache for runtime GPU availability check
_gpu_available_cache = None


def is_gpu_available():
    """Check if a GPU device is available at runtime.

    This function checks whether GPU hardware is actually accessible,
    not just whether muGrid was compiled with GPU support. It attempts
    to create a GPU device and allocate memory to verify the GPU works.

    The result is cached after the first call for performance.

    Returns
    -------
    bool
        True if a GPU device is available and functional, False otherwise.

    Examples
    --------
    >>> import muGrid
    >>> if muGrid.is_gpu_available():
    ...     fc = muGrid.GlobalFieldCollection((10, 10), device=muGrid.Device.gpu())
    ... else:
    ...     fc = muGrid.GlobalFieldCollection((10, 10))

    See Also
    --------
    has_gpu : Compile-time flag indicating GPU support was built.
    """
    global _gpu_available_cache
    if _gpu_available_cache is not None:
        return _gpu_available_cache

    # If not compiled with GPU support, definitely not available
    if not has_gpu:
        _gpu_available_cache = False
        return False

    try:
        # Try to create a GPU device and allocate memory
        # Device.gpu() may succeed even without hardware, so we test allocation
        device = Device.gpu()
        # Use keyword argument for device since positional args differ
        fc = _muGrid.GlobalFieldCollection(
            nb_domain_grid_pts=[2, 2],
            device=device,
        )
        _ = fc.real_field("test", [1])
        _gpu_available_cache = True
        return True
    except Exception:
        _gpu_available_cache = False
        return False

# Import Python wrappers for main classes (these accept wrapped Field objects)
from .Wrappers import (  # noqa: E402, E305
    CartesianDecomposition,
    FEMGradientOperator,
    FFTEngine,
    FileIONetCDF,
    GenericLinearOperator,
    GlobalFieldCollection,
    LaplaceOperator,
    LocalFieldCollection,
)

# Expose OpenMode for FileIONetCDF if available
if hasattr(_muGrid, "FileIONetCDF"):
    OpenMode = _muGrid.FileIONetCDF.OpenMode

# Low-level C++ classes (for advanced use cases)
# These are prefixed with underscore to indicate they're internal
_CartesianDecomposition = _muGrid.CartesianDecomposition
_GenericLinearOperator = _muGrid.GenericLinearOperator
_FFTEngine = _muGrid.FFTEngine
_GlobalFieldCollection = _muGrid.GlobalFieldCollection
_LocalFieldCollection = _muGrid.LocalFieldCollection
if hasattr(_muGrid, "FileIONetCDF"):
    _FileIONetCDF = _muGrid.FileIONetCDF

# Base classes and utilities (always C++ objects)
GradientOperator = _muGrid.GradientOperator
# Backwards compatibility alias
ConvolutionOperatorBase = GradientOperator
Decomposition = _muGrid.Decomposition

# Isotropic stiffness operators (fused elliptic kernels)
IsotropicStiffnessOperator2D = _muGrid.IsotropicStiffnessOperator2D
IsotropicStiffnessOperator3D = _muGrid.IsotropicStiffnessOperator3D
DynCoord = _muGrid.DynCoord
DynRcoord = _muGrid.DynRcoord
IterUnit = _muGrid.IterUnit
Pixel = _muGrid.Pixel
StorageOrder = _muGrid.StorageOrder
SubPt = _muGrid.SubPt
Unit = _muGrid.Unit

# Linear algebra operations on fields (Python wrapper with automatic _cpp extraction)
from . import linalg  # noqa: E402

# FFT utility functions
# Note: fft_freq, fft_freqind, rfft_freq, rfft_freqind are now properties
# on the FFTEngine class (fftfreq, ifftfreq). Use engine.fftfreq instead.
fft_normalization = _muGrid.fft_normalization
get_hermitian_grid_pts = _muGrid.get_hermitian_grid_pts

# Domain indexing utilities
get_domain_ccoord = _muGrid.get_domain_ccoord
get_domain_index = _muGrid.get_domain_index

# Field classes and utilities
from .Field import Field  # noqa: E402
from .Field import wrap_field  # noqa: E402

# MPI communicator and parallel utilities
from .Parallel import Communicator, parprint  # noqa: E402

# Timing utility
from .Timer import Timer  # noqa: E402

# Version information
__version__ = _muGrid.version.description()

# Define public API
__all__ = [
    # Feature flags
    "has_mpi",
    "has_cuda",
    "has_rocm",
    "has_gpu",
    "has_netcdf",
    "host_arch",
    "device_arch",
    # Runtime checks
    "is_gpu_available",
    # Main classes (Python wrappers)
    "CartesianDecomposition",
    "Communicator",
    "GenericLinearOperator",
    "FFTEngine",
    "Field",
    "FileIONetCDF",
    "GlobalFieldCollection",
    "LocalFieldCollection",
    # Field utilities
    "wrap_field",
    # FFT utilities
    "fft_normalization",
    "get_hermitian_grid_pts",
    # Domain utilities
    "get_domain_ccoord",
    "get_domain_index",
    # Enums and types
    "GradientOperator",
    "ConvolutionOperatorBase",  # Backwards compatibility alias
    "LaplaceOperator",
    "FEMGradientOperator",
    "IsotropicStiffnessOperator2D",
    "IsotropicStiffnessOperator3D",
    "Decomposition",
    "DynCoord",
    "DynRcoord",
    "IterUnit",
    "OpenMode",
    "Pixel",
    "StorageOrder",
    "SubPt",
    "Unit",
    "Verbosity",
    # Utilities
    "Timer",
    "parprint",
    # Linear algebra
    "linalg",
]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   Timer.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   25 Dec 2024

@brief  Hierarchical timing utility with nested context manager support

Copyright © 2024 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import json
import time
from contextlib import contextmanager


class Timer:
    """
    Hierarchical timing utility with nested context manager support.

    This class provides fine-grained timing of code sections with support for:
    - Nested timing contexts that track parent-child relationships
    - Accumulation of time across multiple calls to the same timer
    - Call counting for repeated operations
    - Hierarchical summary output in tabular format

    Example usage:
        from muGrid import Timer

        timer = Timer()

        with timer("outer"):
            # some code
            with timer("inner"):
                # nested code
            with timer("inner"):  # called again - time accumulates
                # more nested code

        timer.print_summary()

    Output:
        ==============================================================================
        Timing Summary
        ==============================================================================
        Name                                  Total    Calls      Average   % Parent
        ------------------------------ ------------ -------- ------------ ----------
        outer                             10.00 ms        1            -          -
          inner                            5.00 ms        2     2.50 ms      50.0%
          (other)                          5.00 ms        -            -      50.0%
        ==============================================================================
    """

    def __init__(self, use_papi=False):
        """
        Initialize the timer.

        Args:
            use_papi: Ignored (PAPI support has been removed).
        """
        self._timers = {}  # name -> {"total": float, "calls": int, "children": list}
        self._stack = []  # stack of (name, start_time) for nesting
        self._roots = []  # top-level timer names in order of first use

    @contextmanager
    def __call__(self, name):
        """
        Context manager for timing a named code section.

        Args:
            name: Identifier for this timing section. Repeated calls with the
                  same name accumulate time and increment the call counter.

        Yields:
            None

        Example:
            with timer("my_operation"):
                # code to time
                pass
        """
        # Build hierarchical name based on current stack
        if self._stack:
            parent_name = self._stack[-1][0]
            full_name = f"{parent_name}/{name}"
        else:
            full_name = name

        # Initialize timer if first call
        if full_name not in self._timers:
            self._timers[full_name] = {
                "total": 0.0,
                "calls": 0,
                "children": [],
            }
            # Track as root or as child of parent
            if self._stack:
                parent_full = self._stack[-1][0]
                if full_name not in self._timers[parent_full]["children"]:
                    self._timers[parent_full]["children"].append(full_name)
            else:
                if full_name not in self._roots:
                    self._roots.append(full_name)

        # Push onto stack and start timing
        start = time.perf_counter()
        self._stack.append((full_name, start))

        try:
            yield
        finally:
            # Pop from stack and accumulate time
            elapsed = time.perf_counter() - start
            self._stack.pop()
            self._timers[full_name]["total"] += elapsed
            self._timers[full_name]["calls"] += 1

    def get_time(self, name):
        """
        Get total accumulated time for a timer.

        Args:
            name: Timer name (can be short name or full hierarchical name)

        Returns:
            Total time in seconds, or 0.0 if timer not found
        """
        # Try exact match first, then search for it in hierarchy
        if name in self._timers:
            return self._timers[name]["total"]
        # Search for timer ending with this name
        for full_name in self._timers:
            if full_name.endswith("/" + name) or full_name == name:
                return self._timers[full_name]["total"]
        return 0.0

    def get_calls(self, name):
        """
        Get call count for a timer.

        Args:
            name: Timer name (can be short name or full hierarchical name)

        Returns:
            Number of calls, or 0 if timer not found
        """
        if name in self._timers:
            return self._timers[name]["calls"]
        for full_name in self._timers:
            if full_name.endswith("/" + name) or full_name == name:
                return self._timers[full_name]["calls"]
        return 0

    def reset(self):
        """Clear all timing data."""
        self._timers.clear()
        self._stack.clear()
        self._roots.clear()

    def _format_time(self, seconds):
        """Format time with appropriate units, fixed width (12 chars)."""
        if seconds < 1e-6:
            return f"{seconds * 1e9:9.2f} ns"
        elif seconds < 1e-3:
            return f"{seconds * 1e6:9.2f} us"
        elif seconds < 1:
            return f"{seconds * 1e3:9.2f} ms"
        else:
            return f"{seconds:9.4f} s "

    def _collect_rows(self, name, indent=0, parent_time=None, rows=None):
        """Recursively collect timing data as rows for tabular display."""
        if rows is None:
            rows = []

        info = self._timers[name]
        total = info["total"]
        calls = info["calls"]
        avg = total / calls if calls > 0 else 0

        # Calculate percentage of parent time
        if parent_time and parent_time > 0:
            pct = 100.0 * total / parent_time
        else:
            pct = None

        # Extract short name (last component)
        short_name = name.split("/")[-1]

        # Add row for this timer
        rows.append(
            {
                "indent": indent,
                "name": short_name,
                "total": total,
                "calls": calls,
                "avg": avg if calls > 1 else None,
                "pct": pct,
            }
        )

        # Collect children
        for child in info["children"]:
            self._collect_rows(child, indent + 1, total, rows)

        # Add "other" time if there are children
        if info["children"]:
            children_time = sum(self._timers[c]["total"] for c in info["children"])
            other_time = total - children_time
            if other_time > 1e-9:  # Only show if meaningful
                other_pct = 100.0 * other_time / total if total > 0 else 0
                rows.append(
                    {
                        "indent": indent + 1,
                        "name": "(other)",
                        "total": other_time,
                        "calls": None,
                        "avg": None,
                        "pct": other_pct,
                    }
                )

        return rows

    def print_summary(self, title=None, name_width=30):
        """
        Print hierarchical timing summary in tabular format.

        Args:
            title: Title for the summary section (default: "Timing Summary")
            name_width: Width of the name column (default: 30)
        """
        if not self._timers:
            print("No timing data collected.")
            return

        # Collect all rows
        all_rows = []
        for root in self._roots:
            self._collect_rows(root, rows=all_rows)

        # Determine title
        if title is None:
            title = "Timing Summary"

        line_width = 78
        print(f"\n{'=' * line_width}")
        print(title)
        print(f"{'=' * line_width}")
        print(
            f"{'Name':<{name_width}} {'Total':>12} {'Calls':>8} "
            f"{'Average':>12} {'% Parent':>10}"
        )
        print(f"{'-' * name_width} {'-' * 12} {'-' * 8} {'-' * 12} {'-' * 10}")

        # Print rows
        for row in all_rows:
            # Build indented name
            prefix = "  " * row["indent"]
            name = f"{prefix}{row['name']}"
            if len(name) > name_width:
                name = name[: name_width - 3] + "..."

            # Format columns
            total_str = self._format_time(row["total"])

            if row["calls"] is not None:
                calls_str = f"{row['calls']:>8}"
            else:
                calls_str = f"{'-':>8}"

            if row["pct"] is not None:
                pct_str = f"{row['pct']:>9.1f}%"
            else:
                pct_str = f"{'-':>10}"

            if row["avg"] is not None:
                avg_str = self._format_time(row["avg"])
            else:
                avg_str = f"{'-':>12}"

            print(
                f"{name:<{name_width}} {total_str} {calls_str} {avg_str} {pct_str}"
            )

        print(f"{'=' * line_width}")

    def _build_tree(self, name):
        """Recursively build a tree structure for a timer and its children."""
        info = self._timers[name]
        total = info["total"]
        calls = info["calls"]

        result = {
            "name": name.split("/")[-1],
            "total_seconds": total,
            "calls": calls,
            "avg_seconds": total / calls if calls > 0 else 0,
        }

        if info["children"]:
            result["children"] = [
                self._build_tree(child) for child in info["children"]
            ]
            children_time = sum(self._timers[c]["total"] for c in info["children"])
            other_time = total - children_time
            if other_time > 1e-9:
                result["other_seconds"] = other_time

        return result

    def to_dict(self):
        """
        Return timing data as a hierarchical dictionary.

        Returns:
            Dictionary with hierarchical timer structure suitable for JSON export.
        """
        return {"timers": [self._build_tree(root) for root in self._roots]}

    def to_json(self, indent=2):
        """
        Return timing data as a JSON string.

        Args:
            indent: Indentation level for pretty printing (default: 2)

        Returns:
            JSON string representation of timing data.
        """
        return json.dumps(self.to_dict(), indent=indent)

    def summary_dict(self):
        """
        Return timing data as a flat dictionary for programmatic access.

        Returns:
            Dictionary with timer names as keys and dicts containing
            'total', 'calls', and 'children' as values.
        """
        result = {}
        for name, info in self._timers.items():
            entry = {
                "total": info["total"],
                "calls": info["calls"],
                "avg": info["total"] / info["calls"] if info["calls"] > 0 else 0,
                "children": list(info["children"]),
            }
            result[name] = entry
        return result



import numpy as np

import muGrid

# Grid parameters
nx, ny = 64, 64
Lx, Ly = 2 * np.pi, 2 * np.pi

engine = muGrid.FFTEngine([nx, ny])

# Create fields
f = engine.real_space_field("f")
f_hat = engine.fourier_space_field("f_hat")
grad_x = engine.real_space_field("grad_x")
grad_y = engine.real_space_field("grad_y")
grad_hat = engine.fourier_space_field("grad_hat")

# Get real-space coordinates from engine
X, Y = engine.coords
X = X * Lx  # Scale from [0, 1) to [0, Lx)
Y = Y * Ly

# Initialize with a smooth function: f = sin(x) * cos(y)
f.s[0, 0, :, :] = np.sin(X) * np.cos(Y)

# Forward FFT
engine.fft(f, f_hat)

# Get wavevectors from engine (integer indices, then scale)
# fftfreq returns normalized frequencies in [-0.5, 0.5)
# Multiply by 2*pi*n/L to get physical wavevectors
QX, QY = engine.fftfreq
KX = 2 * np.pi * QX * nx / Lx  # Physical wavevector
KY = 2 * np.pi * QY * ny / Ly

# Derivative in x: multiply by i*kx
grad_hat.s[0, 0, :, :] = 1j * KX * f_hat.s[0, 0, :, :]
engine.ifft(grad_hat, grad_x)
grad_x.s[:] *= engine.normalisation

# Derivative in y: multiply by i*ky
grad_hat.s[0, 0, :, :] = 1j * KY * f_hat.s[0, 0, :, :]
engine.ifft(grad_hat, grad_y)
grad_y.s[:] *= engine.normalisation

# Verify: d/dx[sin(x)cos(y)] = cos(x)cos(y)
expected_grad_x = np.cos(X) * np.cos(Y)
np.testing.assert_allclose(grad_x.s[0, 0], expected_grad_x, atol=1e-10)

print("Fourier derivative test passed!")



"""
Micromechanical homogenization example.

This example computes the effective elastic properties of a periodic volume
element with material microstructure using the linear FEM gradient operator.

The governing equation is the Lippmann-Schwinger equation for linear elasticity:
    K u = -B^T : C : E_macro
where:
    K = B^T : C : B is the stiffness operator
    B is the FEM gradient operator (strain from displacement)
    B^T is the divergence operator (equilibrium)
    C is the material stiffness tensor at each quadrature point
    E_macro is the applied macroscopic strain

After solving, the homogenized stress is computed as:
    Σ = (1/|Ω|) ∫ C : (E_macro + ε(u)) dΩ

For a 2D problem, we compute all 3 independent stiffness components (xx, yy, xy).
For a 3D problem, we compute all 6 independent stiffness components
  (xx, yy, zz, yz, xz, xy).
"""

import argparse
import json

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None

import numpy as np

import muGrid
from muGrid import parprint
from muGrid.Solvers import conjugate_gradients

try:
    from mpi4py import MPI

    comm = muGrid.Communicator(MPI.COMM_WORLD)
except ImportError:
    comm = muGrid.Communicator()

from NuMPI.Testing.Subdivision import suggest_subdivisions

# Voigt notation mappings
# 2D: [xx, yy, xy] -> indices 0, 1, 2
# 3D: [xx, yy, zz, yz, xz, xy] -> indices 0, 1, 2, 3, 4, 5


def voigt_index_2d(i, j):
    """Convert 2D tensor indices to Voigt notation index."""
    if i == j:
        return i
    return 2


def voigt_index_3d(i, j):
    """Convert 3D tensor indices to Voigt notation index.

    Voigt ordering: xx=0, yy=1, zz=2, yz=3, xz=4, xy=5
    """
    if i == j:
        return i
    # Off-diagonal: map (i,j) to Voigt index
    # (1,2) or (2,1) -> yz -> 3
    # (0,2) or (2,0) -> xz -> 4
    # (0,1) or (1,0) -> xy -> 5
    pair = tuple(sorted([i, j]))
    mapping = {(1, 2): 3, (0, 2): 4, (0, 1): 5}
    return mapping[pair]


def voigt_index(dim, i, j):
    """Convert tensor indices to Voigt notation index for given dimension."""
    if dim == 2:
        return voigt_index_2d(i, j)
    else:
        return voigt_index_3d(i, j)


def isotropic_stiffness_2d(E, nu):
    """
    Create 2D plane strain isotropic stiffness tensor in Voigt notation.
    Returns C[3, 3] where [xx, yy, xy] ordering is used.
    """
    lam = E * nu / ((1 + nu) * (1 - 2 * nu))
    mu = E / (2 * (1 + nu))

    C = np.zeros((3, 3))
    C[0, 0] = lam + 2 * mu  # C_xxxx
    C[1, 1] = lam + 2 * mu  # C_yyyy
    C[2, 2] = mu  # C_xyxy
    C[0, 1] = C[1, 0] = lam  # C_xxyy
    return C


def isotropic_stiffness_3d(E, nu):
    """
    Create 3D isotropic stiffness tensor in Voigt notation.
    Returns C[6, 6] where [xx, yy, zz, yz, xz, xy] ordering is used.
    """
    lam = E * nu / ((1 + nu) * (1 - 2 * nu))
    mu = E / (2 * (1 + nu))

    C = np.zeros((6, 6))
    # Diagonal: normal components
    C[0, 0] = lam + 2 * mu  # C_xxxx
    C[1, 1] = lam + 2 * mu  # C_yyyy
    C[2, 2] = lam + 2 * mu  # C_zzzz
    # Diagonal: shear components
    C[3, 3] = mu  # C_yzyz
    C[4, 4] = mu  # C_xzxz
    C[5, 5] = mu  # C_xyxy
    # Off-diagonal: coupling between normal components
    C[0, 1] = C[1, 0] = lam  # C_xxyy
    C[0, 2] = C[2, 0] = lam  # C_xxzz
    C[1, 2] = C[2, 1] = lam  # C_yyzz
    return C


def isotropic_stiffness(dim, E, nu):
    """Create isotropic stiffness tensor for given dimension."""
    if dim == 2:
        return isotropic_stiffness_2d(E, nu)
    else:
        return isotropic_stiffness_3d(E, nu)


def create_microstructure(coords, inclusion_type="single", inclusion_radius=0.25):
    """
    Create a phase field (0=matrix, 1=inclusion) based on coordinates.
    """
    dim = len(coords)

    if inclusion_type == "single":
        center = [0.5] * dim
        r2 = sum((c - center[i]) ** 2 for i, c in enumerate(coords))
        phase = np.where(r2 < inclusion_radius**2, 1.0, 0.0)
    elif inclusion_type == "checkerboard":
        if dim == 2:
            x, y = coords
            phase = ((x > 0.5).astype(float) + (y > 0.5).astype(float)) % 2
        else:
            x, y, z = coords
            phase = (
                (x > 0.5).astype(float)
                + (y > 0.5).astype(float)
                + (z > 0.5).astype(float)
            ) % 2
    elif inclusion_type == "random":
        np.random.seed(42)
        shape = coords[0].shape
        phase = (np.random.random(shape) < 0.3).astype(float)
    else:
        raise ValueError(f"Unknown inclusion type: {inclusion_type}")

    return phase


parser = argparse.ArgumentParser(
    prog="Homogenization",
    description="Compute effective elastic properties via FEM-based homogenization",
)

parser.add_argument(
    "-n",
    "--nb-grid-pts",
    default=[16, 16],
    type=lambda s: [int(x) for x in s.split(",")],
    help="Grid points as nx,ny or nx,ny,nz (default: 16,16)",
)

_devices = {
    "cpu": muGrid.Device.cpu(),
    "gpu": muGrid.Device.gpu(),  # Auto-detect CUDA or ROCm
}

parser.add_argument(
    "-d",
    "--device",
    choices=_devices,
    default="cpu",
    help="Device for computation: 'cpu' or 'gpu' (auto-detect CUDA/ROCm) "
    "(default: cpu)",
)

parser.add_argument(
    "-i",
    "--maxiter",
    type=int,
    default=500,
    help="Maximum number of CG iterations (default: 500)",
)

parser.add_argument(
    "-t",
    "--tol",
    type=float,
    default=1e-6,
    help="CG convergence tolerance (default: 1e-6)",
)

parser.add_argument(
    "--E-matrix",
    type=float,
    default=1.0,
    help="Young's modulus of matrix (default: 1.0)",
)

parser.add_argument(
    "--E-inclusion",
    type=float,
    default=10.0,
    help="Young's modulus of inclusion (default: 10.0)",
)

parser.add_argument(
    "--nu",
    type=float,
    default=0.3,
    help="Poisson's ratio (same for both phases, default: 0.3)",
)

parser.add_argument(
    "--inclusion-type",
    choices=["single", "checkerboard", "random"],
    default="single",
    help="Type of inclusion pattern (default: single)",
)

parser.add_argument(
    "--inclusion-radius",
    type=float,
    default=0.25,
    help="Radius of inclusion for 'single' type (default: 0.25)",
)

parser.add_argument(
    "-p",
    "--plot",
    action="store_true",
    help="Show plot of microstructure and stress fields (default: off, 2D only)",
)

parser.add_argument(
    "-q",
    "--quiet",
    action="store_true",
    help="Suppress per-iteration output (default: off)",
)

parser.add_argument(
    "--json",
    action="store_true",
    help="Output results in JSON format (implies --quiet)",
)

parser.add_argument(
    "-k",
    "--kernel",
    choices=["fused", "generic"],
    default="fused",
    help="Kernel implementation: 'fused' (IsotropicStiffnessOperator) or "
    "'generic' (gradient/stress/divergence) (default: fused)",
)

args = parser.parse_args()

# JSON implies quiet mode
if args.json:
    args.quiet = True

# Select array library based on memory location
if args.device == "cpu":
    import numpy as arr
else:
    import cupy as arr

device = _devices[args.device]

# Parse grid dimensions
dim = len(args.nb_grid_pts)
if dim not in (2, 3):
    raise ValueError("Only 2D and 3D grids are supported")

# Physical domain size (unit cell)
domain_size = np.ones(dim)
grid_spacing = domain_size / np.array(args.nb_grid_pts)

# Number of Voigt components
nb_voigt = 3 if dim == 2 else 6  # 2D: xx, yy, xy; 3D: xx, yy, zz, yz, xz, xy

# Voigt component labels for output
if dim == 2:
    voigt_labels = ["xx", "yy", "xy"]
else:
    voigt_labels = ["xx", "yy", "zz", "yz", "xz", "xy"]

# Create the FEM gradient operator first to get accurate quadrature info
gradient_op = muGrid.FEMGradientOperator(dim, tuple(grid_spacing))

# Number of quadrature points from the operator
# Number of nodes per element: 4 for 2D (corners of pixel), 8 for 3D (corners of voxel)
nb_quad = gradient_op.nb_quad_pts
nb_nodes = 2**dim

# Quadrature weights (area/volume of each element)
quad_weights = np.array(gradient_op.quadrature_weights)

# Determine MPI decomposition using NuMPI's suggest_subdivisions
s = suggest_subdivisions(dim, comm.size)

# Create Cartesian decomposition for ghost handling.
# The FEM gradient kernel requires ghosts for accessing neighbor nodes.
# CartesianDecomposition handles ghost communication for both serial and
# MPI parallel execution.
#
# We use ghosts on BOTH sides (left and right). This approach means:
# - Ghost elements on the left boundary provide contributions to interior nodes
#   directly
# - Ghost elements on the right boundary provide contributions to interior nodes
#   directly
# - No ghost reduction is needed since interior nodes receive all contributions
#   directly
# - The trade-off is slightly more memory and computation in ghost regions
decomposition = muGrid.CartesianDecomposition(
    comm,
    args.nb_grid_pts,
    nb_subdivisions=s,  # MPI-aware domain decomposition
    nb_ghosts_left=(1,) * dim,
    nb_ghosts_right=(1,) * dim,
    nb_sub_pts={"quad": nb_quad},  # Use actual quad count from operator
    device=device,
)

# Get local grid dimensions from the decomposition
# For MPI runs, this is the local subdomain shape, not the global grid
local_grid_shape = decomposition.nb_subdomain_grid_pts

# Get coordinates for microstructure generation from the decomposition
# This returns only the local portion of coordinates for this MPI rank
# Shape: [dim, *local_grid_shape] with values in [0, 1)
local_coords = decomposition.coords
coords = [local_coords[d] for d in range(dim)]

# Create the microstructure (phase field) - only for local portion
phase = create_microstructure(coords, args.inclusion_type, args.inclusion_radius)

# Create the material stiffness tensor at each pixel
C_matrix = isotropic_stiffness(dim, args.E_matrix, args.nu)
C_inclusion = isotropic_stiffness(dim, args.E_inclusion, args.nu)

if not args.quiet:
    parprint(f"Grid size: {args.nb_grid_pts}", comm=comm)
    parprint(f"Dimensions: {dim}D", comm=comm)
    parprint(f"Grid spacing: {grid_spacing}", comm=comm)
    parprint(f"Device: {device.device_string}", comm=comm)
    parprint(f"Number of quadrature points per pixel: {nb_quad}", comm=comm)
    parprint(f"Number of nodal points per pixel: {nb_nodes}", comm=comm)
    parprint(f"Quadrature weights: {quad_weights}", comm=comm)
    # Compute global volume fraction via MPI reduction
    local_sum = np.sum(phase)
    local_count = phase.size
    global_sum = comm.sum(local_sum)
    global_count = comm.sum(local_count)
    v_f_print = global_sum / global_count
    parprint(f"Inclusion volume fraction: {v_f_print:.4f}", comm=comm)
    parprint(comm=comm)

# Create muGrid fields for gradient operations.
# The FEM gradient operator now supports multi-component fields:
# - Input: vector field with (dim,) components -> shape (dim, sub_pts, pixels)
# - Output: tensor field with (dim, dim) components at quad pts
#   (dim input components × dim operators)
#
# Tensor fields for gradient/stress at quadrature points:
grad_u = decomposition.real_field(
    "grad_u", (dim, dim), "quad"
)  # displacement gradient tensor
stress_field = decomposition.real_field(
    "stress_field", (dim, dim), "quad"
)  # stress tensor

# Vector fields for CG solver (displacement and force vectors)
u_field = decomposition.real_field("u_field", (dim,))
f_field = decomposition.real_field("f_field", (dim,))
rhs_field = decomposition.real_field("rhs_field", (dim,))

# Material stiffness at each quadrature point [voigt, voigt, quad, *local_grid_shape]
# Create on host first, then convert to device array if needed
# phase is already local (from decomposition.coords), so no slicing needed
C_field_shape = (nb_voigt, nb_voigt, nb_quad) + local_grid_shape
C_field_np = np.zeros(C_field_shape)
for q in range(nb_quad):
    for i in range(nb_voigt):
        for j in range(nb_voigt):
            C_field_np[i, j, q] = (
                C_matrix[i, j] * (1 - phase) + C_inclusion[i, j] * phase
            )

# Convert to device array if using GPU
C_field = arr.asarray(C_field_np)

# For fused kernel: create Lamé parameter fields
if args.kernel == "fused":
    # Compute Lamé parameters from Young's modulus and Poisson's ratio
    lam_matrix = args.E_matrix * args.nu / ((1 + args.nu) * (1 - 2 * args.nu))
    mu_matrix = args.E_matrix / (2 * (1 + args.nu))
    lam_inclusion = args.E_inclusion * args.nu / ((1 + args.nu) * (1 - 2 * args.nu))
    mu_inclusion = args.E_inclusion / (2 * (1 + args.nu))

    # Create element-based CartesianDecomposition with ghost cells
    element_decomposition = muGrid.CartesianDecomposition(
        comm,
        args.nb_grid_pts,
        nb_subdivisions=s,  # Use same MPI decomposition
        nb_ghosts_left=(1,) * dim,
        nb_ghosts_right=(1,) * dim,
        device=device,
    )

    # Create lambda and mu fields
    lambda_field = element_decomposition.real_field("lambda")
    mu_field = element_decomposition.real_field("mu")

    # Set spatially varying material properties based on phase
    # Phase is defined at grid points, element properties taken at grid points
    # (for element (i,j), use material at grid point (i,j))
    # phase is already local (from decomposition.coords), so no slicing needed
    lambda_field.p[...] = arr.asarray(lam_matrix * (1 - phase) + lam_inclusion * phase)
    mu_field.p[...] = arr.asarray(mu_matrix * (1 - phase) + mu_inclusion * phase)

    # Fill ghost cells for material fields (only needs to be done once)
    element_decomposition.communicate_ghosts(lambda_field)
    element_decomposition.communicate_ghosts(mu_field)

    # Create the fused isotropic stiffness operator
    if dim == 2:
        fused_stiffness_op = muGrid.IsotropicStiffnessOperator2D(tuple(grid_spacing))
    else:
        fused_stiffness_op = muGrid.IsotropicStiffnessOperator3D(tuple(grid_spacing))

    if not args.quiet:
        parprint(f"Using fused IsotropicStiffnessOperator{dim}D", comm=comm)
        parprint(f"  Lambda (matrix): {lam_matrix:.4f}, "
                 f"Mu (matrix): {mu_matrix:.4f}", comm=comm)
        parprint(
            f"  Lambda (incl.):  {lam_inclusion:.4f}, Mu (incl.):  {mu_inclusion:.4f}",
            comm=comm,
        )

# Create global timer for hierarchical timing
timer = muGrid.Timer()

# Performance counters
nb_grid_pts_total = np.prod(args.nb_grid_pts)


def compute_strain(u_vec, strain_out):
    """
    Compute strain from displacement field.

    The gradient operator directly handles vector input:
    - Input: vector field u with (dim,) components
    - Output: tensor field ∂u_i/∂x_j with (dim, dim) components
    - Strain: ε_ij = 0.5 * (∂u_i/∂x_j + ∂u_j/∂x_i)

    Parameters
    ----------
    u_vec : Field
        Vector displacement field with shape (dim, nb_nodes, pixels)
    strain_out : ndarray
        Output strain array with shape (dim, dim, quad, *grid_shape)
    """
    with timer("communicate_ghosts"):
        # Fill ghost values for periodic BC
        decomposition.communicate_ghosts(u_vec)

    with timer("gradient_apply"):
        # Compute gradient tensor: grad_u.s[i, j, ...] = ∂u_i/∂x_j
        gradient_op.apply(u_vec, grad_u)

    # Compute symmetric strain: ε_ij = 0.5 * (∂u_i/∂x_j + ∂u_j/∂x_i)
    grad = grad_u.s
    strain_out[...] = 0.5 * (grad + grad.swapaxes(0, 1))


def strain_to_voigt(strain):
    """
    Convert strain tensor to Voigt notation.

    Parameters
    ----------
    strain : ndarray
        Strain tensor with shape (dim, dim, quad, *grid_shape)

    Returns
    -------
    eps_voigt : ndarray
        Strain in Voigt notation with shape (nb_voigt, quad, *grid_shape)
    """
    voigt_shape = (nb_voigt, nb_quad) + local_grid_shape
    eps_voigt = arr.zeros(voigt_shape)

    if dim == 2:
        eps_voigt[0, ...] = strain[0, 0, ...]  # exx
        eps_voigt[1, ...] = strain[1, 1, ...]  # eyy
        eps_voigt[2, ...] = 2 * strain[0, 1, ...]  # 2*exy (engineering shear)
    else:  # dim == 3
        eps_voigt[0, ...] = strain[0, 0, ...]  # exx
        eps_voigt[1, ...] = strain[1, 1, ...]  # eyy
        eps_voigt[2, ...] = strain[2, 2, ...]  # ezz
        eps_voigt[3, ...] = 2 * strain[1, 2, ...]  # 2*eyz (engineering shear)
        eps_voigt[4, ...] = 2 * strain[0, 2, ...]  # 2*exz (engineering shear)
        eps_voigt[5, ...] = 2 * strain[0, 1, ...]  # 2*exy (engineering shear)

    return eps_voigt


def voigt_to_stress(sig_voigt, stress):
    """
    Convert stress from Voigt notation to tensor.

    Parameters
    ----------
    sig_voigt : ndarray
        Stress in Voigt notation with shape (nb_voigt, quad, *grid_shape)
    stress : ndarray
        Output stress tensor with shape (dim, dim, quad, *grid_shape)
    """
    if dim == 2:
        stress[0, 0, ...] = sig_voigt[0, ...]  # sxx
        stress[1, 1, ...] = sig_voigt[1, ...]  # syy
        stress[0, 1, ...] = sig_voigt[2, ...]  # sxy
        stress[1, 0, ...] = sig_voigt[2, ...]  # syx
    else:  # dim == 3
        stress[0, 0, ...] = sig_voigt[0, ...]  # sxx
        stress[1, 1, ...] = sig_voigt[1, ...]  # syy
        stress[2, 2, ...] = sig_voigt[2, ...]  # szz
        stress[1, 2, ...] = sig_voigt[3, ...]  # syz
        stress[2, 1, ...] = sig_voigt[3, ...]  # szy
        stress[0, 2, ...] = sig_voigt[4, ...]  # sxz
        stress[2, 0, ...] = sig_voigt[4, ...]  # szx
        stress[0, 1, ...] = sig_voigt[5, ...]  # sxy
        stress[1, 0, ...] = sig_voigt[5, ...]  # syx


def compute_stress(strain, stress, C):
    """
    Compute stress from strain using Voigt notation.

    Parameters
    ----------
    strain : ndarray
        Strain tensor with shape (dim, dim, quad, *grid_shape)
    stress : ndarray
        Output stress tensor with shape (dim, dim, quad, *grid_shape)
    C : ndarray
        Material stiffness in Voigt notation (nb_voigt, nb_voigt, quad, *grid_shape)
    """
    with timer("compute_stress"):
        # Convert strain to Voigt notation
        eps_voigt = strain_to_voigt(strain)

        # Compute stress in Voigt: sig = C @ eps
        if device.is_host:
            sig_voigt = np.einsum("ijq...,jq...->iq...", C, eps_voigt)
        else:
            # CuPy einsum
            sig_voigt = arr.einsum("ijq...,jq...->iq...", C, eps_voigt)

        # Convert back to tensor
        voigt_to_stress(sig_voigt, stress)


def compute_divergence(stress, f_vec):
    """
    Compute divergence of stress tensor.

    The transpose of the gradient operator directly handles tensor input:
    - Input: stress tensor with (dim, dim) components
    - Output: force vector with (dim,) components
    - f_i = Σ_j ∂σ_ij/∂x_j (divergence of stress)

    With two-sided ghosts:
    1. We fill ghost pixel stresses via communicate_ghosts (periodic copies)
    2. The transpose reads stress from ALL pixels (interior + ghost)
    3. Ghost elements contribute directly to interior nodes
    4. No ghost reduction needed

    Parameters
    ----------
    stress : ndarray
        Stress tensor with shape (dim, dim, quad, *grid_shape)
    f_vec : Field
        Output vector force field with shape (dim, nb_nodes, *grid_shape)
    """
    # Copy stress to field and fill ghost values
    stress_field.s[...] = stress
    with timer("communicate_ghosts"):
        decomposition.communicate_ghosts(stress_field)

    with timer("gradient_transpose"):
        # Apply transpose (divergence) with quadrature weights
        # The transpose sums over operators (j direction) for each input component (i)
        f_vec.pg[...] = 0.0
        gradient_op.transpose(stress_field, f_vec, list(quad_weights))


# Temporary arrays for strain and stress [dim, dim, quad, *local_grid_shape]
strain_shape = (dim, dim, nb_quad) + local_grid_shape
strain_arr = arr.zeros(strain_shape)
stress_arr = arr.zeros(strain_shape)


def apply_stiffness_generic(u_in, f_out):
    """
    Apply K = B^T C B to displacement vector using generic gradient/stress/divergence.

    Parameters
    ----------
    u_in : Field
        Input displacement field with (dim, nb_nodes) components
    f_out : Field
        Output force field with (dim, nb_nodes) components (modified in place)
    """
    with timer("apply_stiffness"):
        with timer("strain"):
            # Compute strain eps = B * u
            compute_strain(u_in, strain_arr)

        with timer("stress"):
            # Compute stress sig = C : eps
            compute_stress(strain_arr, stress_arr, C_field)

        with timer("divergence"):
            # Compute force f = B^T * sig
            compute_divergence(stress_arr, f_out)


def apply_stiffness_fused(u_in, f_out):
    """
    Apply K = B^T C B to displacement vector using fused isotropic stiffness operator.

    This is more memory-efficient for isotropic materials as it only stores
    two Lamé parameters (λ, μ) per element instead of the full stiffness tensor.

    Parameters
    ----------
    u_in : Field
        Input displacement field with (dim, nb_nodes) components
    f_out : Field
        Output force field with (dim, nb_nodes) components (modified in place)
    """
    with timer("apply_stiffness"):
        with timer("communicate_ghosts"):
            # Fill ghost values for periodic BC
            decomposition.communicate_ghosts(u_in)

        with timer("fused_kernel"):
            # Apply fused stiffness operator: f = K @ u
            fused_stiffness_op.apply(u_in, lambda_field, mu_field, f_out)


# Select the stiffness application method
if args.kernel == "fused":
    apply_stiffness = apply_stiffness_fused
else:
    apply_stiffness = apply_stiffness_generic


def compute_rhs(E_macro, rhs_out):
    """
    Compute RHS: f = -B^T C E_macro

    Parameters
    ----------
    E_macro : ndarray
        Macroscopic strain tensor [dim, dim]
    rhs_out : Field
        Output field for RHS (modified in place)
    """
    # Create uniform strain field from macroscopic strain
    eps_macro = arr.zeros(strain_shape)
    for i in range(dim):
        for j in range(dim):
            eps_macro[i, j, ...] = E_macro[i, j]

    # Compute stress from macroscopic strain
    sig_macro = arr.zeros_like(eps_macro)
    compute_stress(eps_macro, sig_macro, C_field)

    # Compute divergence (with negative sign for RHS)
    compute_divergence(sig_macro, rhs_out)
    rhs_out.s[...] *= -1.0


# Storage for homogenized stiffness
C_eff = np.zeros((nb_voigt, nb_voigt))

# Macroscopic strain cases
if dim == 2:
    strain_cases = [
        (0, 0),  # xx
        (1, 1),  # yy
        (0, 1),  # xy
    ]
else:  # dim == 3
    strain_cases = [
        (0, 0),  # xx
        (1, 1),  # yy
        (2, 2),  # zz
        (1, 2),  # yz
        (0, 2),  # xz
        (0, 1),  # xy
    ]

if not args.quiet:
    parprint("=" * 60, comm=comm)
    parprint("Computing homogenized stiffness tensor", comm=comm)
    parprint("=" * 60, comm=comm)

total_iterations = 0

with timer("total_solve"):
    for case_idx, (i, j) in enumerate(strain_cases):
        # Create unit macroscopic strain
        E_macro = arr.zeros((dim, dim))
        E_macro[i, j] = 1.0
        if i != j:
            E_macro[j, i] = 1.0  # Symmetric

        voigt_col = voigt_index(dim, i, j)

        if not args.quiet:
            parprint(f"\nCase {case_idx + 1}: E_macro[{i},{j}] = 1", comm=comm)

        # Compute RHS into rhs_field
        compute_rhs(E_macro, rhs_field)

        # Initialize displacement to zero
        u_field.s[...] = 0.0

        # CG callback
        iteration_count = [0]  # Use list to allow modification in nested function

        def callback(iteration, state):
            iteration_count[0] = iteration
            if not args.quiet and iteration % 10 == 0:
                res_norm = float(arr.sqrt(state["rr"]))
                parprint(f"  CG iteration {iteration}: |r| = {res_norm:.6e}", comm=comm)

        # Solve K u = f using conjugate_gradients from Solvers.py
        try:
            conjugate_gradients(
                comm,
                decomposition,
                rhs_field,
                u_field,
                hessp=apply_stiffness,
                tol=args.tol,
                callback=callback,
                maxiter=args.maxiter,
                timer=timer,
            )
            converged = True
        except RuntimeError as e:
            if "did not converge" in str(e):
                converged = False
            else:
                raise

        total_iterations += iteration_count[0]

        if not args.quiet:
            if converged:
                parprint(f"  CG converged in {iteration_count[0]} "
                         "iterations", comm=comm)
            else:
                parprint(f"  CG did not converge after {args.maxiter} "
                         "iterations", comm=comm)

        # Compute strain from solution
        compute_strain(u_field, strain_arr)

        # Add macroscopic strain to get total strain
        for ii in range(dim):
            for jj in range(dim):
                strain_arr[ii, jj, ...] += E_macro[ii, jj]

        # Compute stress from total strain
        compute_stress(strain_arr, stress_arr, C_field)

        # Compute average stress (homogenized stress)
        # Σ_kl = (1/V) ∫ σ_kl dV = (1/V) Σ_q w_q * σ_kl(q)
        sig_avg = np.zeros((dim, dim))
        for k in range(dim):
            for L in range(dim):
                local_sum = 0.0
                for q in range(nb_quad):
                    if device.is_host:
                        local_sum += quad_weights[q] * np.sum(stress_arr[k, L, q, ...])
                    else:
                        local_sum += quad_weights[q] * float(
                            arr.sum(stress_arr[k, L, q, ...])
                        )
                sig_avg[k, L] = comm.sum(local_sum)

        # Normalize by total volume
        total_volume = np.prod(domain_size)
        sig_avg /= total_volume

        # Store in homogenized stiffness (column voigt_col)
        for k in range(dim):
            for L in range(dim):
                voigt_row = voigt_index(dim, k, L)
                # Only store unique components (upper triangle in tensor, all in Voigt)
                if k <= L:
                    C_eff[voigt_row, voigt_col] = sig_avg[k, L]

        if not args.quiet:
            if dim == 2:
                parprint(
                    f"  Average stress: xx={sig_avg[0, 0]:.6f}, "
                    f"yy={sig_avg[1, 1]:.6f}, xy={sig_avg[0, 1]:.6f}",
                    comm=comm,
                )
            else:
                parprint(
                    f"  Average stress: xx={sig_avg[0, 0]:.6f}, "
                    f"yy={sig_avg[1, 1]:.6f}, zz={sig_avg[2, 2]:.6f}",
                    comm=comm,
                )
                parprint(
                    f"                  yz={sig_avg[1, 2]:.6f}, "
                    f"xz={sig_avg[0, 2]:.6f}, xy={sig_avg[0, 1]:.6f}",
                    comm=comm,
                )

# Get timing information
elapsed_time = timer.get_time("total_solve")
# Stiffness is called in both startup and iteration phases
nb_stiffness_calls_startup = timer.get_calls("total_solve/startup/apply_stiffness")
nb_stiffness_calls_iter = timer.get_calls("total_solve/iteration/hessp/apply_stiffness")
nb_stiffness_calls = nb_stiffness_calls_startup + nb_stiffness_calls_iter
apply_stiffness_time_startup = timer.get_time("total_solve/startup/apply_stiffness")
apply_stiffness_time_iter = timer.get_time(
    "total_solve/iteration/hessp/apply_stiffness"
)
apply_stiffness_time = apply_stiffness_time_startup + apply_stiffness_time_iter

# Lattice updates per second (LUPS)
# One "lattice update" = one grid point processed in one stiffness call
total_lattice_updates = nb_grid_pts_total * nb_stiffness_calls
lups = total_lattice_updates / elapsed_time if elapsed_time > 0 else 0
apply_lups = (
    total_lattice_updates / apply_stiffness_time if apply_stiffness_time > 0 else 0
)

# Performance metrics
# Memory throughput estimate:
# - FEM gradient: reads neighbor values, writes gradient
# - Stress computation: reads strain, writes stress
# - Divergence: reads stress, writes force
# Approximate: 2 * dim * dim * nb_quad + dim values per grid point, 8 bytes each
bytes_per_call = nb_grid_pts_total * (2 * dim * dim * nb_quad + dim) * 8
total_bytes = nb_stiffness_calls * bytes_per_call
memory_throughput = total_bytes / elapsed_time if elapsed_time > 0 else 0

# FLOPS estimate (rough)
# - Gradient: ~dim * dim * nb_stencil_pts * 2 FLOPs per grid point
# - Stress: ~nb_voigt * nb_voigt * nb_quad FLOPs per grid point
# - Divergence: similar to gradient
flops_per_call = nb_grid_pts_total * (
    dim * dim * 10 + nb_voigt * nb_voigt * nb_quad * 2
)
total_flops = nb_stiffness_calls * flops_per_call
flops_rate = total_flops / elapsed_time if elapsed_time > 0 else 0

# Analytical bounds
# Compute global volume fraction via MPI reduction
local_phase_sum = np.sum(phase)
local_phase_count = phase.size
global_phase_sum = comm.sum(local_phase_sum)
global_phase_count = comm.sum(local_phase_count)
v_f = float(global_phase_sum / global_phase_count)  # Volume fraction of inclusion
E_m, E_i = args.E_matrix, args.E_inclusion
nu = args.nu
E_voigt = v_f * E_i + (1 - v_f) * E_m
E_reuss = 1.0 / (v_f / E_i + (1 - v_f) / E_m)
E_eff_approx = C_eff[0, 0] * (1 - nu**2)

if args.json:
    # JSON output
    # Timer's to_dict() includes PAPI data when available
    results = {
        "config": {
            "nb_grid_pts": [int(x) for x in args.nb_grid_pts],
            "nb_grid_pts_total": int(nb_grid_pts_total),
            "dimensions": int(dim),
            "device": device.device_string,
            "maxiter": int(args.maxiter),
            "tolerance": float(args.tol),
            "E_matrix": float(args.E_matrix),
            "E_inclusion": float(args.E_inclusion),
            "nu": float(args.nu),
            "inclusion_type": args.inclusion_type,
            "inclusion_radius": float(args.inclusion_radius),
            "volume_fraction": float(v_f),
            "kernel": args.kernel,
        },
        "results": {
            "total_cg_iterations": int(total_iterations),
            "stiffness_calls": int(nb_stiffness_calls),
            "total_time_seconds": float(elapsed_time),
            "apply_stiffness_time_seconds": float(apply_stiffness_time),
            "total_lattice_updates": int(total_lattice_updates),
            "MLUPS": float(lups / 1e6),
            "GLUPS": float(lups / 1e9),
            "apply_MLUPS": float(apply_lups / 1e6),
            "apply_GLUPS": float(apply_lups / 1e9),
            "bytes_per_call": int(bytes_per_call),
            "total_bytes": int(total_bytes),
            "memory_throughput_GBps": float(memory_throughput / 1e9),
            "flops_per_call_estimated": int(flops_per_call),
            "total_flops_estimated": int(total_flops),
            "flops_rate_GFLOPs_estimated": float(flops_rate / 1e9),
            "C_eff": [
                [float(C_eff[i, j]) for j in range(nb_voigt)] for i in range(nb_voigt)
            ],
            "E_effective_approx": float(E_eff_approx),
            "E_voigt_bound": float(E_voigt),
            "E_reuss_bound": float(E_reuss),
        },
        "timing": timer.to_dict(),
    }
    parprint(json.dumps(results, indent=2), comm=comm)
else:
    # Text output
    parprint("\n" + "=" * 60, comm=comm)
    parprint("Homogenized stiffness tensor (Voigt notation)", comm=comm)
    parprint("=" * 60, comm=comm)

    # Print header
    header = "      " + "".join(f"{lbl:>10}" for lbl in voigt_labels)
    parprint(header, comm=comm)

    # Print matrix
    for i, row_label in enumerate(voigt_labels):
        row = f"{row_label:>4}  " + "".join(
            f"{C_eff[i, j]:10.6f}" for j in range(nb_voigt)
        )
        parprint(row, comm=comm)

    parprint("\n" + "=" * 60, comm=comm)
    parprint("Comparison with analytical bounds", comm=comm)
    parprint("=" * 60, comm=comm)
    parprint(f"Volume fraction of inclusion: {v_f:.4f}", comm=comm)
    parprint(f"Voigt bound (upper): E = {E_voigt:.4f}", comm=comm)
    parprint(f"Reuss bound (lower): E = {E_reuss:.4f}", comm=comm)
    parprint(f"Effective E (approx from C_xxxx): E ≈ {E_eff_approx:.4f}", comm=comm)

    parprint("\n" + "=" * 60, comm=comm)
    parprint("Performance Summary", comm=comm)
    parprint("=" * 60, comm=comm)
    parprint(
        f"Grid size: {' x '.join(map(str, args.nb_grid_pts))} = "
        f"{nb_grid_pts_total:,} points",
        comm=comm,
    )
    parprint(f"Dimensions: {dim}D", comm=comm)
    parprint(f"Device: {device.device_string}", comm=comm)
    parprint(f"Total CG iterations: {total_iterations}", comm=comm)
    parprint(f"Stiffness operator calls: {nb_stiffness_calls}", comm=comm)
    parprint(f"Total time: {elapsed_time:.4f} seconds", comm=comm)

    parprint("\nLattice updates per second:", comm=comm)
    parprint(f"  Total lattice updates: {total_lattice_updates:,}", comm=comm)
    parprint(f"  LUPS: {lups / 1e6:.2f} MLUPS ({lups / 1e9:.4f} GLUPS)", comm=comm)

    parprint("\nMemory throughput (estimated):", comm=comm)
    parprint(f"  Bytes per stiffness call: {bytes_per_call / 1e6:.2f} MB", comm=comm)
    parprint(f"  Total bytes transferred: {total_bytes / 1e9:.2f} GB", comm=comm)
    parprint(f"  Throughput: {memory_throughput / 1e9:.2f} GB/s", comm=comm)

    parprint("\nFLOPS (estimated):", comm=comm)
    parprint(f"  FLOPs per stiffness call: {flops_per_call / 1e6:.2f} MFLOP", comm=comm)
    parprint(f"  Total FLOPs: {total_flops / 1e9:.2f} GFLOP", comm=comm)
    parprint(f"  FLOP rate: {flops_rate / 1e9:.2f} GFLOP/s", comm=comm)
    parprint("=" * 60, comm=comm)

    # Print hierarchical timing breakdown (includes PAPI data when enabled)
    timer.print_summary()

# Optional plotting (2D only)
if args.plot:
    if dim == 3:
        parprint("Warning: Plotting not supported for 3D grids", comm=comm)
    elif plt is None:
        parprint("Warning: matplotlib not available, cannot show plot", comm=comm)
    elif comm.rank == 0:
        fig, axes = plt.subplots(1, 3, figsize=(12, 4))

        # Microstructure
        ax = axes[0]
        im = ax.imshow(phase.T, origin="lower", cmap="viridis")
        ax.set_title("Microstructure (0=matrix, 1=inclusion)")
        plt.colorbar(im, ax=ax)

        # Stress xx from last load case (convert to numpy if on device)
        ax = axes[1]
        if device.is_host:
            sig_xx_avg = np.mean(stress_arr[0, 0, ...], axis=0)
        else:
            sig_xx_avg = np.mean(arr.asnumpy(stress_arr[0, 0, ...]), axis=0)
        im = ax.imshow(sig_xx_avg.T, origin="lower", cmap="RdBu_r")
        ax.set_title(r"$\sigma_{xx}$ (last load case)")
        plt.colorbar(im, ax=ax)

        # Displacement magnitude from last load case
        ax = axes[2]
        if device.is_host:
            u_mag = np.sqrt(u_field.s[0, 0, ...] ** 2 + u_field.s[1, 0, ...] ** 2)
        else:
            u_s = arr.asnumpy(u_field.s)
            u_mag = np.sqrt(u_s[0, 0, ...] ** 2 + u_s[1, 0, ...] ** 2)
        im = ax.imshow(u_mag.T, origin="lower", cmap="viridis")
        ax.set_title("|u| (last load case)")
        plt.colorbar(im, ax=ax)

        plt.tight_layout()
        plt.show()



"""
@file    python_poisson_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    10 April 2025

@brief   solving the Poisson equation as a compound test

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np
import pytest
from NuMPI.Testing.Subdivision import suggest_subdivisions

import muGrid
from muGrid import wrap_field
from muGrid.Solvers import conjugate_gradients


def test_fd_stencil():
    stencil = np.array(
        [[0, 1, 0], [1, -4, 1], [0, 1, 0]]
    )  # FD-stencil for the Laplacian
    laplace = muGrid.GenericLinearOperator([-1, -1], stencil)
    assert laplace.nb_output_components == 1
    assert laplace.nb_quad_pts == 1

    nb_ghosts = (1, 1)
    fc = muGrid.GlobalFieldCollection(
        [3, 3], nb_ghosts_left=nb_ghosts, nb_ghosts_right=nb_ghosts
    )
    ifield = fc.real_field("input-field")
    ofield = fc.real_field("output-field")
    ifield.p[...] = 1
    ifield.p[1, 1] = 2
    # Manually correct a periodic boundary
    ifield.pg[...] = np.pad(ifield.p, tuple(zip(nb_ghosts, nb_ghosts)), mode="wrap")
    laplace.apply(ifield, ofield)
    np.testing.assert_allclose(ofield.p, stencil)
    np.testing.assert_allclose(np.sum(ifield.p * ofield.p), -4)


def test_fd_poisson_solver(comm, nb_grid_pts=(128, 128)):
    """Finite-differences Poisson solver"""
    s = suggest_subdivisions(len(nb_grid_pts), comm.size)

    decomposition = muGrid.CartesianDecomposition(comm, nb_grid_pts, s, (1, 1), (1, 1))
    grid_spacing = 1 / np.array(nb_grid_pts)  # Grid spacing

    stencil = np.array(
        [[0, 1, 0], [1, -4, 1], [0, 1, 0]]
    )  # FD-stencil for the Laplacian
    laplace = muGrid.GenericLinearOperator([-1, -1], stencil)
    assert laplace.nb_output_components == 1
    assert laplace.nb_quad_pts == 1

    # np.testing.assert_array_equal(decomposition.nb_subdivisions, s)

    x, y = decomposition.coords  # Domain-local coords for each pixel

    rhs = decomposition.real_field("rhs")
    solution = decomposition.real_field("solution")

    rhs.p[...] = np.sin(2 * np.pi * x)

    def callback(iteration, state):
        """
        Callback function to print the iteration and squared residual norm.
        """
        print(iteration, state["rr"])

    def hessp(x_field, Ax_field):
        """
        Function to compute the product of the Hessian matrix with a vector.
        The Hessian is represented by the convolution operator.
        """
        decomposition.communicate_ghosts(x_field)
        laplace.apply(x_field, Ax_field)
        # We need the minus sign because the Laplace operator is negative
        # definite, but the conjugate-gradients solver assumes a
        # positive-definite operator.
        Ax_field.s[...] /= -np.mean(grid_spacing) ** 2  # Scale by grid spacing

    conjugate_gradients(
        comm,
        decomposition,
        rhs,
        solution,
        hessp=hessp,
        tol=1e-6,
        callback=callback,
        maxiter=10,
    )

    np.testing.assert_allclose(
        solution.p, np.sin(2 * np.pi * x) / (2 * np.pi) ** 2, atol=1e-5
    )


@pytest.mark.skip("Currently fails; reenable after migration")
def test_unit_impulse(
    comm,
):
    nx, ny = nb_grid_pts = (4, 6)  # grid size should be arbitrary
    s = suggest_subdivisions(len(nb_grid_pts), comm.size)
    decomposition = muGrid.CartesianDecomposition(comm, nb_grid_pts, s, (1, 1), (1, 1))
    fc = decomposition._cpp.collection
    fc.set_nb_sub_pts("quad_points", 2)
    fc.set_nb_sub_pts("nodal_points", 1)

    # set up gradient operator
    gradient = np.array(
        [
            [  # Derivative in x-direction
                [[[-1, 0], [1, 0]]],  # Bottom-left triangle (first quadrature point)
                [[[0, -1], [0, 1]]],  # Top-right triangle (second quadrature point)
            ],
            [  # Derivative in y-direction
                [[[-1, 1], [0, 0]]],  # Bottom-left triangle (first quadrature point)
                [[[0, 0], [-1, 1]]],  # Top-right triangle (second quadrature point)
            ],
        ],
    )
    gradient_op = muGrid.GenericLinearOperator([0, 0], gradient)

    # Get nodal field
    nodal_field_cpp = fc.real_field("nodal-field", (1,), "nodal_points")
    impuls_response_field_cpp = fc.real_field(
        "impuls_response_field", (1,), "nodal_points"
    )
    nodal_field = wrap_field(nodal_field_cpp)
    impuls_response_field = wrap_field(impuls_response_field_cpp)

    # Get quadrature field of shape (2, quad, nx, ny)
    quad_field_cpp = fc.real_field("quad-field", (2,), "quad_points")

    # set up impulse
    impuls_locations = (impuls_response_field.icoordsg[0] == 0) & (
        impuls_response_field.icoordsg[1] == 0
    )

    left_location = (impuls_response_field.icoordsg[0] == nx - 1) & (
        impuls_response_field.icoordsg[1] == 0
    )
    right_location = (impuls_response_field.icoordsg[0] == 1) & (
        impuls_response_field.icoordsg[1] == 0
    )
    top_location = (impuls_response_field.icoordsg[0] == 0) & (
        impuls_response_field.icoordsg[1] == 1
    )
    bottom_location = (impuls_response_field.icoordsg[0] == 0) & (
        impuls_response_field.icoordsg[1] == ny - 1
    )

    nodal_field.sg[0, 0, impuls_locations] = 1
    impuls_response_field.sg[0, 0, impuls_locations] = 4
    impuls_response_field.sg[0, 0, left_location] = -1
    impuls_response_field.sg[0, 0, right_location] = -1
    impuls_response_field.sg[0, 0, top_location] = -1
    impuls_response_field.sg[0, 0, bottom_location] = -1

    decomposition.communicate_ghosts(nodal_field)
    print(
        f"unit impuls: nodal field with buffers in rank {comm.rank} \n "
        + f"{nodal_field.sg}"
    )
    print(
        f"impuls_response_field: nodal field with buffers in rank {comm.rank} \n "
        + f"{impuls_response_field.sg}"
    )

    # Apply the gradient operator to the nodal field and write result to the quad field
    gradient_op.apply(nodal_field, wrap_field(quad_field_cpp))

    decomposition.communicate_ghosts(wrap_field(quad_field_cpp))

    # Apply the gradient transposed operator to the quad field and write result to
    # the nodal field
    gradient_op.transpose(
        quadrature_point_field=wrap_field(quad_field_cpp),
        nodal_field=nodal_field,
        weights=[
            1 / 2,
            1 / 2,
        ],  # size of the element is half of the pixel. Pixel size is 1
    )

    decomposition.communicate_ghosts(nodal_field)
    print(
        "computed unit impuls response: nodal field with buffers in rank "
        f"{comm.rank} \n {nodal_field.sg}"
    )

    print(
        f"local sum on core (nodal_field.s) = {np.sum(nodal_field.s)}"
    )  # does not have to be zero
    local_sum = np.sum(nodal_field.s)
    total_sum = comm.sum(local_sum)
    print(f"total_sum= {total_sum}")  # have to be zero

    # Check that the nodal_field has zero mean
    np.testing.assert_allclose(
        total_sum,
        0,
        atol=1e-10,
    )
    # Check that the impulse response is correct
    np.testing.assert_allclose(
        nodal_field.s,
        impuls_response_field.s,
        atol=1e-5,
    )



"""
Collection of simple parallel solvers
"""

from . import linalg


def conjugate_gradients(
    comm,
    fc,
    b,
    x,
    hessp: callable,
    tol: float = 1e-6,
    maxiter: int = 1000,
    callback: callable = None,
    timer=None,
):
    """
    Conjugate gradient method for matrix-free solution of the linear problem
    Ax = b, where A is represented by the function hessp (which computes the
    product of A with a vector). The method iteratively refines the solution x
    until the residual ||Ax - b|| is less than tol or until maxiter iterations
    are reached.

    Parameters
    ----------
    comm : muGrid.Communicator
        Communicator for parallel processing.
    fc : muGrid.GlobalFieldCollection, muGrid.LocalFieldCollection, or
         muGrid.CartesianDecomposition
        Collection for creating temporary fields used by the CG algorithm.
    b : muGrid.Field
        Right-hand side vector.
    x : muGrid.Field
        Initial guess for the solution (modified in place).
    hessp : callable
        Function that computes the product of the Hessian matrix A with a vector.
        Signature: hessp(input_field, output_field) where both are muGrid.Field.
    tol : float, optional
        Tolerance for convergence. The default is 1e-6.
    maxiter : int, optional
        Maximum number of iterations. The default is 1000.
    callback : callable, optional
        Function to call after each iteration with signature:
        callback(iteration, state_dict) where state_dict contains:
        - "x": solution field
        - "r": residual field
        - "p": search direction field
        - "rr": squared residual norm (float)
    timer : muGrid.Timer, optional
        Timer object for performance profiling. If provided, the solver will
        record timing for the hessp (Hessian-vector product) operations.

    Returns
    -------
    x : muGrid.Field
        Solution to the system Ax = b (same as input field x).

    Raises
    ------
    RuntimeError
        If the algorithm does not converge within maxiter iterations,
        or if the residual becomes NaN (indicating numerical issues).
    """
    tol_sq = tol * tol

    # Timer context manager (no-op if timer is None)
    from contextlib import nullcontext

    def timed(name):
        return timer(name) if timer is not None else nullcontext()

    with timed("startup"):
        # Create temporary fields with matching component shape
        # r: residual field
        # p: search direction field
        # Ap: Hessian product field
        r = fc.real_field("cg-residual", b.components_shape)
        p = fc.real_field("cg-search-direction", b.components_shape)
        Ap = fc.real_field("cg-hessian-product", b.components_shape)

        # Initial residual: r = b - A*x
        hessp(x, Ap)
        # r = b (copy)
        linalg.copy(b, r)
        # r = r - Ap = b - A*x (axpy with alpha=-1)
        linalg.axpy(-1.0, Ap, r)

        # Initial search direction: p = r
        linalg.copy(r, p)

        if callback:
            rr = comm.sum(linalg.norm_sq(r))
            callback(0, {"x": x, "r": r, "p": p, "rr": rr})

        with timed("dot_rr"):
            rr = comm.sum(linalg.norm_sq(r))
        rr_val = float(rr)

        if rr_val < tol_sq:
            return x

    with timed("iteration"):
        for iteration in range(maxiter):
            # Compute Hessian product: Ap = A * p
            with timed("hessp"):
                hessp(p, Ap)

            # Compute pAp for step size
            with timed("dot_pAp"):
                pAp = comm.sum(linalg.vecdot(p, Ap))

            # Compute alpha
            alpha = rr / pAp

            # Update solution: x += alpha * p
            with timed("update_x"):
                linalg.axpy(alpha, p, x)

            # Update residual and compute norm in one pass: r -= alpha * Ap
            with timed("update_r"):
                next_rr = comm.sum(linalg.axpy_norm_sq(-alpha, Ap, r))
            next_rr_val = float(next_rr)

            if callback:
                with timed("callback"):
                    callback(iteration + 1, {"x": x, "r": r, "p": p, "rr": next_rr})

            # Check for numerical issues (NaN indicates non-positive-definite H)
            if next_rr_val != next_rr_val:  # NaN check
                raise RuntimeError(
                    "Residual became NaN - Hessian may not be positive definite"
                )

            if next_rr_val < tol_sq:
                return x

            # Compute beta
            beta = next_rr / rr

            # Update rr for next iteration
            rr = next_rr

            # Update search direction: p = r + beta * p
            with timed("update_p"):
                linalg.axpby(1.0, r, beta, p)

    raise RuntimeError("Conjugate gradient algorithm did not converge")



import numpy as np
import muGrid

if not muGrid.has_netcdf:
    print("IO example is skipped: muGrid was built without NetCDF support")
    exit(0)

# Two dimensional grid
nb_grid_pts = (11, 12, 13)
fc = muGrid.GlobalFieldCollection(nb_grid_pts, nb_sub_pts={"element": 5})

# Get a tensor-field (for example to represent the strain)
strain = fc.real_field("strain", (3, 3), "element")

# Fill the field with random numbers
strain.s[...] = np.random.rand(*((3, 3, 5) + nb_grid_pts))

# Initialize a file I/O object (using string for open mode)
file = muGrid.FileIONetCDF("example.nc", open_mode="overwrite")
file.register_field_collection(fc)  # Register the field collection with the file
file.append_frame().write()  # Write all fields of the field collection to the file
